from unittest import TestCase
from unittest.mock import patch

import numpy as np
from pytest import approx

from btb.selection.pure import PureBestKVelocity


class TestPureBestKVelocity(TestCase):
    # METHOD: __init__(self, choices, **kwargs)
    # VALIDATE:
    #     * attribute values

    def test___init__(self):

        # Run
        selector = PureBestKVelocity(['RF', 'SVM'], k=3)

        # Assert
        assert selector.choices == ['RF', 'SVM']
        assert selector.k == 3

    # METHOD: compute_rewards(self, scores)
    # VALIDATE:
    #     * returned values
    # NOTES:
    #     * This is 1to1 with BestKVelocity. Why is this not inheriting from it?

    def test_compute_rewards_kt_k(self):
        """Less scores than self.k: No padding"""

        # Set-up
        selector = PureBestKVelocity(['RF', 'SVM'], k=5)

        # Run
        scores = [0.5, 0.6, 0.75, 0.8]
        rewards = selector.compute_rewards(scores)

        # Assert
        np.testing.assert_allclose(rewards, [0.05, 0.15, 0.1])

    def test_compute_rewards_gt_k(self):
        """More scores than self.k: padding"""

        # Set-up
        selector = PureBestKVelocity(['RF', 'SVM'], k=3)

        # Run
        scores = [0.1, 0.5, 0.6, 0.75, 0.8]
        rewards = selector.compute_rewards(scores)

        # Assert
        np.testing.assert_allclose(rewards, [0.05, 0.15, 0.1, 0., 0.])

    # METHOD: select(self, choice_scores)
    # VALIDATE:
    #     * Returned values for multiple cases
    #     * arguments passed to bandit
    # TODO:
    #     * This code can be simplified to be more easy to understand

    @patch('btb.selection.pure.PureBestKVelocity.bandit')
    def test_select_more_scores_than_k_min(self, bandit_mock):
        """If min length is gt k_min, self.compute_rewards is used.

        In this case, we expect the lower scores to be zeroed.
        """

        # Set-up
        selector = PureBestKVelocity(['RF', 'SVM'])

        bandit_mock.return_value = 'RF'

        # Run
        choice_scores = {
            'DT': [0.7, 0.75, 0.73],
            'RF': [0.8, 0.85, 0.83],
            'SVM': [0.9, 0.95, 0.91],
        }
        best = selector.select(choice_scores)

        # Assert
        assert best == 'RF'

        choice_rewards = {
            'RF': [approx(0.02), approx(0.03)],
            'SVM': [approx(0.04), approx(0.01)],
        }
        bandit_mock.assert_called_once_with(choice_rewards)

    @patch('btb.selection.pure.PureBestKVelocity.bandit')
    def test_select_less_scores_than_k_min(self, bandit_mock):
        """If min length is lt k_min, super().compute_rewards is used.

        In this case, we expect the socres to be returned as they are.
        """

        # Set-up
        selector = PureBestKVelocity(['RF', 'SVM'])

        bandit_mock.return_value = 'RF'

        # Run
        choice_scores = {
            'DT': [0.7, 0.75, 0.73],
            'RF': [0.8],
            'SVM': [0.9],
        }
        best = selector.select(choice_scores)

        # Assert
        assert best == 'RF'

        choice_rewards = {
            'RF': [1],
            'SVM': [1],
        }
        bandit_mock.assert_called_once_with(choice_rewards)



from unittest import TestCase
from unittest.mock import patch

from btb.selection.hierarchical import HierarchicalByAlgorithm


class TestHierarchicalByAlgorithm(TestCase):
    # METHOD: __init__(self, choices, **kwargs)
    # VALIDATE:
    #     * attribute values
    # TODO:
    #     * Why not re-use choices instead of creating an extra argument?

    def test___init__(self):

        # Run
        by_algorithm = {
            'DT': frozenset(('DT', 'RF', 'ET')),
            'SVM': frozenset(('LSVC', 'NuSVC')),
        }
        selector = HierarchicalByAlgorithm(choices=None, by_algorithm=by_algorithm)

        # Assert
        assert selector.choices is None
        assert selector.by_algorithm == by_algorithm

    # METHOD: select(self, choice_scores)
    # VALIDATE:
    #     * returned value
    #     * arguments passed to self.bandit
    # NOTES:
    #     * self.bandit will need to be mocked to capture the arguments
    #     * Why is a new object created? Why not set self.choices and use super.select?
    #     * Why are frozensets needed? Why not tuples?
    # TODO:
    #     * "if not condition: continue" should rather be "if condition: do something"

    @patch('btb.selection.hierarchical.HierarchicalByAlgorithm.bandit')
    def test_select(self, bandit_mock):
        """If min length is lt k_min, super().compute_rewards is used.

        In this case, we expect the socres to be returned as they are.
        """

        # Set-up
        by_algorithm = {
            'DT': ('DT', 'RF', 'ET'),    # We use tuples here to
            'SVM': ('LSVC', 'NuSVC'),    # preserve the order.
            'KNN': ('KNN', ),
        }
        selector = HierarchicalByAlgorithm(choices=None, by_algorithm=by_algorithm)

        bandit_mock.return_value = 'SVM'

        # Run
        choice_scores = {
            'DT': [0.7, 0.8, 0.9],
            'RF': [0.94],
            'LSVC': [0.88, 0.95, 0.93],
            'NuSVC': [0.89, 0.91, 0.92],
        }
        best = selector.select(choice_scores)

        # Assert
        assert best == 'LSVC'

        alg_scores = {
            'DT': [0.7, 0.8, 0.9, 0.94],
            'SVM': [0.88, 0.95, 0.93, 0.89, 0.91, 0.92]
        }
        bandit_mock.assert_called_once_with(alg_scores)



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock

from sklearn.gaussian_process import GaussianProcessRegressor

from btb.tuning.tunable import Tunable
from btb.tuning.tuners.gaussian_process import GPEiTuner, GPTuner


class TestGaussianProcessTuner(TestCase):
    """Test GaussianProcessTuner class."""

    def test___init__(self):
        # setup
        tunable = MagicMock(spec_set=Tunable)
        # run
        instance = GPTuner(tunable)

        # assert
        assert instance._MODEL_CLASS == GaussianProcessRegressor
        assert instance._metamodel_kwargs == {'length_scale': 0.1}


class TestGaussianProcessExpectedImprovementTuner(TestCase):
    """Test GaussianProcessExpectedImprovementTuner class."""

    def test___init__(self):
        # setup
        tunable = MagicMock(spec_set=Tunable)

        # run
        instance = GPEiTuner(tunable)

        # assert
        assert instance._MODEL_CLASS == GaussianProcessRegressor
        assert instance._metamodel_kwargs == {'length_scale': 0.1}



# -*- coding: utf-8 -*-

"""Top level of the tuning module."""

from btb.tuning.hyperparams.boolean import BooleanHyperParam
from btb.tuning.hyperparams.categorical import CategoricalHyperParam
from btb.tuning.hyperparams.numerical import FloatHyperParam, IntHyperParam
from btb.tuning.tunable import Tunable
from btb.tuning.tuners.base import StopTuning
from btb.tuning.tuners.gaussian_process import GPEiTuner, GPTuner
from btb.tuning.tuners.uniform import UniformTuner

__all__ = (
    'BooleanHyperParam',
    'CategoricalHyperParam',
    'GPEiTuner',
    'GPTuner',
    'FloatHyperParam',
    'IntHyperParam',
    'StopTuning',
    'Tunable',
    'UniformTuner',
)



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock

import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor

from btb.tuning.metamodels.gaussian_process import GaussianProcessMetaModel


class TestGaussianProcessMetaModel(TestCase):

    def test___init__(self):
        # run
        instance = GaussianProcessMetaModel()

        # assert
        assert instance._MODEL_KWARGS_DEFAULT == {'normalize_y': True}
        assert instance._MODEL_CLASS == GaussianProcessRegressor

    def test__predict(self):
        # setup
        instance = MagicMock()
        instance._model_instance.predict.return_value = [[1], [2]]

        # run
        result = GaussianProcessMetaModel._predict(instance, 1)

        # assert
        instance._model_instance.predict.assert_called_once_with(1, return_std=True)
        np.testing.assert_array_equal(result, np.array([[1, 2]]))



from unittest import TestCase
from unittest.mock import patch

import numpy as np

from btb.selection.best import BestKReward, BestKVelocity


class TestBestKReward(TestCase):

    # METHOD: __init__(self, choices, **kwargs)
    # VALIDATE:
    #     * attribute values

    def test___init__(self):

        # Run
        selector = BestKReward(['RF', 'SVM'], k=3)

        # Assert
        assert selector.choices == ['RF', 'SVM']
        assert selector.k == 3

    # METHOD: compute_rewards(self, scores)
    # VALIDATE:
    #     * returned values
    #     * check the case whete len(scores) < self.k

    def test_compute_rewards_lt_k(self):

        # Set-up
        selector = BestKReward(['RF', 'SVM'], k=3)

        # Run
        scores = [0.8, 0.9]
        rewards = selector.compute_rewards(scores)

        # Assert
        assert rewards == scores

    def test_compute_rewards_gt_k(self):

        # Set-up
        selector = BestKReward(['RF', 'SVM'], k=3)

        # Run
        scores = [0.83, 0.8, 0.86, 0.9]
        rewards = selector.compute_rewards(scores)

        # Assert
        np.testing.assert_array_equal(rewards, [0.83, np.nan, 0.86, 0.9])

    def test_compute_rewards_duplicates(self):
        k = 3
        choices = ['RF', 'SVM']
        scores = [0.7, 0.8, 0.7, 0.1, 0.8, 0.7]
        selector = BestKReward(choices, k=k)
        rewards = selector.compute_rewards(scores)
        np.testing.assert_array_equal(
            np.sort(rewards),
            np.sort([np.nan, np.nan, np.nan, 0.7, 0.8, 0.8])
        )

    # METHOD: select(self, choice_scores)
    # VALIDATE:
    #     * Returned values for multiple cases
    #     * arguments passed to bandit
    # TODO:
    #     * This code can be simplified to be more easy to understand
    # NOTES:
    #     * Should this be >= self.k instead of >= K_MIN?

    @patch('btb.selection.best.BestKReward.bandit')
    def test_select_more_scores_than_k_min(self, bandit_mock):
        """If min length is gt k_min, self.compute_rewards is used.

        In this case, we expect the lower scores to be nan.
        """

        # Set-up
        selector = BestKReward(['RF', 'SVM'])

        bandit_mock.return_value = 'SVM'

        # Run
        choice_scores = {
            'DT': [0.7, 0.75, 0.73],
            'RF': [0.8, 0.85, 0.83],
            'SVM': [0.9, 0.95, 0.93],
        }
        best = selector.select(choice_scores)

        # Assert
        assert best == 'SVM'

        choice_rewards_expected = {
            'RF': [np.nan, 0.85, 0.83],
            'SVM': [np.nan, 0.95, 0.93],
        }
        # TODO
        (choice_rewards, ), _ = bandit_mock.call_args
        assert choice_rewards.keys() == choice_rewards_expected.keys()
        for k in choice_rewards:
            assert k in choice_rewards_expected
            np.testing.assert_array_equal(
                choice_rewards[k],
                choice_rewards_expected[k]
            )

    @patch('btb.selection.best.BestKReward.bandit')
    def test_select_less_scores_than_k_min(self, bandit_mock):
        """If min length is lt k_min, super().compute_rewards is used.

        In this case, we expect the socres to be returned as they are.
        """

        # Set-up
        selector = BestKReward(['RF', 'SVM'])

        bandit_mock.return_value = 'SVM'

        # Run
        choice_scores = {
            'DT': [0.7, 0.75, 0.73],
            'RF': [0.8],
            'SVM': [0.9, 0.95, 0.93],
        }
        best = selector.select(choice_scores)

        # Assert
        assert best == 'SVM'

        choice_rewards = {
            'RF': [0.8],
            'SVM': [0.9, 0.95, 0.93],
        }
        bandit_mock.assert_called_once_with(choice_rewards)


class TestBestKVelocity(TestCase):

    # METHOD: compute_rewards(self, scores)
    # VALIDATE:
    #     * returned values

    def test_compute_rewards_kt_k(self):
        """Less scores than self.k: No padding"""

        # Set-up
        selector = BestKVelocity(['RF', 'SVM'], k=5)

        # Run
        scores = [0.5, 0.6, 0.75, 0.8]
        rewards = selector.compute_rewards(scores)

        # Assert
        np.testing.assert_allclose(
            np.sort(rewards),
            np.sort([0.05, 0.15, 0.1])
        )

    def test_compute_rewards_gt_k(self):
        """More scores than self.k: padding"""

        # Set-up
        selector = BestKVelocity(['RF', 'SVM'], k=3)

        # Run
        scores = [0.1, 0.5, 0.6, 0.75, 0.8]
        rewards = selector.compute_rewards(scores)

        # Assert
        np.testing.assert_allclose(
            np.sort(rewards),
            np.sort([0.05, 0.15, 0.1, np.nan, np.nan])
        )



from btb.selection.best import BestKReward, BestKVelocity
from btb.selection.hierarchical import HierarchicalByAlgorithm
from btb.selection.pure import PureBestKVelocity
from btb.selection.recent import RecentKReward, RecentKVelocity
from btb.selection.ucb1 import UCB1
from btb.selection.uniform import Uniform

__all__ = (
    'BestKReward', 'BestKVelocity', 'HierarchicalByAlgorithm',
    'PureBestKVelocity', 'RecentKReward', 'RecentKVelocity',
    'UCB1', 'Uniform',
)



# -*- coding: utf-8 -*-

from collections import Counter, defaultdict
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

import numpy as np
from tqdm.autonotebook import trange

from btb.session import BTBSession
from btb.tuning.tuners.base import StopTuning
from btb.tuning.tuners.gaussian_process import GPTuner


class TestBTBSession(TestCase):

    def test__normalize_maximize_true(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)

        # run
        result = BTBSession._normalize(instance, 1)

        # assert
        assert result == 1

    def test__normalize_maximize_false(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance._maximize = False

        # run
        result = BTBSession._normalize(instance, 1)

        # assert
        assert result == -1

    def test___init__default(self):
        # run
        tunable = {'my_test_tuner': {'my_tunable_hp': {}}}
        scorer = 'my_scorer'
        instance = BTBSession(tunable, scorer)

        # assert
        assert instance._tunables is tunable
        assert instance._scorer is scorer
        assert instance._tuner_class is GPTuner
        assert instance._best_normalized == -np.inf
        assert instance._normalized_scores == defaultdict(list)
        assert instance._tuners == {}
        assert instance._tunable_names == ['my_test_tuner']
        assert instance._range is range
        assert instance._max_errors == 1
        assert instance._maximize

        assert instance.best_score is None
        assert instance.best_proposal is None
        assert instance.proposals == {}
        assert instance.iterations == 0
        assert instance.errors == Counter()

    def test___init__custom(self):
        # setup
        selector = MagicMock()

        # run
        tunable = {'my_test_tuner': {'my_tunable_hp': {}}}
        scorer = 'my_scorer'
        instance = BTBSession(
            tunable,
            scorer,
            tuner_class='my_tuner',
            selector_class=selector,
            maximize=False,
            max_errors=2,
            verbose=True
        )

        # assert
        assert instance._tunables is tunable
        assert instance._scorer is scorer
        assert instance._tuner_class == 'my_tuner'
        assert instance._max_errors == 2
        assert instance._best_normalized == -np.inf
        assert instance._normalized_scores == defaultdict(list)
        assert instance._tuners == {}
        assert instance._tunable_names == ['my_test_tuner']
        assert instance._range is trange

        assert instance.best_proposal is None
        assert instance.proposals == {}
        assert instance.iterations == 0
        assert instance.errors == Counter()

    def test__make_dumpable(self):
        # run
        randint = np.random.randint(1, dtype=np.integer)
        to_dump = {
            1: randint,
            'str': 'None',
            'array': np.array([1, 2, 3]),
        }

        result = BTBSession._make_dumpable(MagicMock(), to_dump)

        # assert
        expected_result = {
            '1': int(randint),
            'str': None,
            'array': [1, 2, 3]
        }

        assert result == expected_result

    @patch('btb.session.np.random.choice')
    def test__get_next_tunable_name_normalized_scores(self, mock_np_random_choice):
        # setup
        mock_np_random_choice.return_value = 'test_name'
        selector = MagicMock()
        selector.select.return_value = 'test_name'

        instance = MagicMock(spec_set=BTBSession)
        instance._normalized_scores = [('test_name', 0.1), ('second_test_name', 0.2)]
        instance._selector = selector

        # run
        tunable_name = BTBSession._get_next_tunable_name(instance)

        # assert
        assert tunable_name == 'test_name'
        mock_np_random_choice.assert_not_called()
        selector.select.assert_called_once_with([
            ('test_name', 0.1),
            ('second_test_name', 0.2),
        ])

    @patch('btb.session.np.random.choice')
    def test__get_next_tunable_name_normalized_scores_none(self, mock_np_random_choice):
        # setup
        mock_np_random_choice.return_value = 'test_name'

        instance = MagicMock(spec_set=BTBSession)
        instance._normalized_scores = None
        instance._tunables = {
            'test_name': 'test',
            'second_test_name': 'second_test',
        }

        # python3.5 randomness issue, we read as in `btb.session`
        expected_mock_call = list(instance._tunables.keys())

        # run
        tunable_name = BTBSession._get_next_tunable_name(instance)

        # assert

        assert tunable_name == 'test_name'
        mock_np_random_choice.assert_called_once_with(expected_mock_call)

    def test_propose_no_tunables(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance._tunables = None

        # run
        with self.assertRaises(StopTuning):
            BTBSession.propose(instance)

    @patch('btb.session.isinstance')
    @patch('btb.session.Tunable')
    def test_propose_normalized_scores_lt_tunable_names(self, mock_tunable, mock_isinstance):
        # setup
        mock_tunable.from_dict.return_value.get_defaults.return_value = 'defaults'
        mock_isinstance.return_value = True

        tuner = MagicMock()

        instance = MagicMock(spec_set=BTBSession)
        instance._tuner_class = tuner
        instance.proposals = {}
        instance._normalized_scores.__len__.return_value = 0
        instance._tunables = {'test_tunable': 'test_spec'}
        instance._tunable_names = ['test_tunable']

        instance._make_id.return_value = 1

        # run
        res_name, res_config = BTBSession.propose(instance)

        # assert
        assert res_name == 'test_tunable'
        assert res_config == 'defaults'

        expected_proposals = {
            1: {
                'id': 1,
                'name': 'test_tunable',
                'config': 'defaults'
            }
        }
        assert instance.proposals == expected_proposals

        instance._make_id.assert_called_once_with('test_tunable', 'defaults')
        mock_tunable.from_dict.assert_called_once_with('test_spec')
        tuner.assert_called_once_with(mock_tunable.from_dict.return_value)
        mock_tunable.from_dict.return_value.get_defaults.assert_called_once_with()

        expected_isinstance_calls = [call('test_spec', dict), call('defaults', mock_tunable)]
        mock_isinstance.has_calls(expected_isinstance_calls)

    def test_propose_normalized_scores_gt_tunable_names(self):
        # setup
        tuner = MagicMock()
        tuner.propose.return_value = 'parameters'

        instance = MagicMock(spec_set=BTBSession)
        instance.proposals = {}
        instance._normalized_scores.__len__.return_value = 1

        instance._selector.select.return_value = 'test_tunable'
        instance._tuners = {'test_tunable': tuner}
        instance._tunables = {'test_tunable': 'test_spec'}
        instance._tunable_names = ['test_tunable']
        instance._get_next_tunable_name.return_value = 'test_tunable'

        instance._make_id.return_value = 1

        # run
        res_name, res_config = BTBSession.propose(instance)

        # assert
        assert res_name == 'test_tunable'
        assert res_config == 'parameters'

        expected_proposals = {
            1: {
                'id': 1,
                'name': 'test_tunable',
                'config': 'parameters'
            }
        }
        assert instance.proposals == expected_proposals

        instance._make_id.assert_called_once_with('test_tunable', 'parameters')
        tuner.propose.assert_called_once_with(1)

    def test_propose_raise_error(self):
        # setup

        tuner = MagicMock()
        tuner.propose.side_effect = [StopTuning('test')]

        instance = MagicMock(spec_set=BTBSession)
        instance._normalized_scores.__len__.return_value = 1

        instance._selector.select.return_value = 'test_tunable'
        instance._tuners = {'test_tunable': tuner}
        instance._tunables = {'test_tunable': 'test_spec'}
        instance._tunable_names = ['test_tunable']
        instance._get_next_tunable_name.return_value = 'test_tunable'

        instance._make_id.return_value = 1

        # run
        with self.assertRaises(ValueError):
            BTBSession.propose(instance)

    @patch('btb.session.isinstance')
    @patch('btb.session.Tunable')
    def test_propose_tunable_cardinality_eq_one(self, mock_tunable, mock_isinstance):
        # setup
        mock_tunable.from_dict.return_value.cardinality = 1
        mock_tunable.from_dict.return_value.get_defaults.return_value = 'parameters'
        mock_isinstance.return_value = True

        instance = MagicMock(spec_set=BTBSession)
        instance._tuners = {}
        instance._tunable_names = ['test_tunable']
        instance.proposals = {}

        instance._make_id.return_value = 1

        # run
        tunable_name, config = BTBSession.propose(instance)

        # assert
        instance._make_id.assert_called_once_with('test_tunable', 'parameters')
        instance._tuner_class.assert_not_called()

        assert instance._tuners == {'test_tunable': None}
        assert 'test_tunable' == tunable_name
        assert 'parameters' == config

    def test_propose_tuner_is_none(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance._tuners = {'test_tunable': None}
        instance._tunable_names = ['test_tunable']
        instance._normalized_scores = None
        instance._get_next_tunable_name.return_value = 'test_tunable'

        # run
        with self.assertRaises(ValueError):
            BTBSession.propose(instance)

    def test_handle_error_errors_lt_max_errors(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance.errors = Counter()
        instance._max_errors = 2

        # run
        BTBSession.handle_error(instance, 'test')

        # assert
        instance._normalized_scores.pop.assert_not_called()
        instance._tunable_names.remove.assert_not_called()

    def test_handle_error_errors_gt_max_errors(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance.errors = Counter()
        instance._max_errors = 0

        # run
        BTBSession.handle_error(instance, 'test')

        # assert
        instance._remove_tunable.assert_called_once_with('test')

    def test_record_score_is_none(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance._make_id.return_value = 0
        instance.proposals = [{'test': 'test'}]
        instance.errors = Counter()
        instance._max_errors = 5

        # run
        BTBSession.record(instance, 'test', 'config', None)

        # assert
        instance.handle_error.assert_called_once_with('test')

    def test_record_score_gt_best(self):
        # setup
        tuner = MagicMock()

        instance = MagicMock(spec_set=BTBSession)
        instance._make_id.return_value = 0
        instance.proposals = [{'test': 'test'}]
        instance._tuners = {'test': tuner}
        instance.best_proposal = None

        instance._best_normalized = 0
        instance._normalize.return_value = 1
        instance._normalized_scores = defaultdict(list)

        # run
        BTBSession.record(instance, 'test', 'config', 1)

        # assert
        expected_normalized_scores = defaultdict(list)
        expected_normalized_scores['test'].append(1)

        assert instance._normalized_scores == expected_normalized_scores
        assert instance.best_proposal == {'test': 'test', 'score': 1}
        assert instance._best_normalized == 1

        tuner.record.assert_called_once_with('config', 1)

    def test_record_score_gt_best_tuner_none(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance._make_id.return_value = 0
        instance.proposals = [{'test': 'test'}]
        instance._tuners = {'test': None}
        instance.best_proposal = None

        instance._best_normalized = 0
        instance._normalize.return_value = 1
        instance._normalized_scores = defaultdict(list)

        # run
        BTBSession.record(instance, 'test', 'config', 1)

        # assert
        expected_normalized_scores = defaultdict(list)
        expected_normalized_scores['test'].append(1)

        assert instance._normalized_scores == expected_normalized_scores
        assert instance.best_proposal == {'test': 'test', 'score': 1}
        assert instance._best_normalized == 1

    def test_record_score_lt_best(self):
        # setup
        tuner = MagicMock()

        instance = MagicMock(spec_set=BTBSession)
        instance._make_id.return_value = 0
        instance.proposals = [{'test': 'test'}]
        instance._tuners = {'test': tuner}
        instance.best_proposal = None

        instance._best_normalized = 10
        instance._normalize.return_value = 1
        instance._normalized_scores = defaultdict(list)

        # run
        BTBSession.record(instance, 'test', 'config', 1)

        # assert
        expected_normalized_scores = defaultdict(list)
        expected_normalized_scores['test'].append(1)

        assert instance.best_proposal is None
        assert instance._normalized_scores == expected_normalized_scores
        assert instance._best_normalized == 10

        tuner.record.assert_called_once_with('config', 1)

    def test_run_score(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance.propose.return_value = ('test', 'config')
        instance._scorer.return_value = 1
        instance.best_proposal = {'test': 'config'}
        instance._range = range
        instance.iterations = 0

        # run
        result = BTBSession.run(instance, 1)

        # assert
        instance._scorer.assert_called_once_with('test', 'config')
        instance.record.assert_called_once_with('test', 'config', 1)
        assert result == {'test': 'config'}
        assert instance.iterations == 1

    def test_run_score_none(self):
        # setup
        instance = MagicMock(spec_set=BTBSession)
        instance.propose.return_value = ('test', {'hp': 'test'})
        instance._scorer.side_effect = Exception()
        instance.best_proposal = {'test': 'config'}
        instance._range = range
        instance.iterations = 0

        # run
        result = BTBSession.run(instance, 1)

        # assert
        instance._scorer.assert_called_once_with('test', {'hp': 'test'})
        instance.record.assert_called_once_with('test', {'hp': 'test'}, None)
        assert result == {'test': 'config'}
        assert instance.iterations == 1



# -*- coding: utf-8 -*-

import random

from btb.tuning import GPTuner, Tunable
from btb.tuning.hyperparams import (
    BooleanHyperParam, CategoricalHyperParam, FloatHyperParam, IntHyperParam)


def test_tuning():
    hyperparams = {
        'bhp': BooleanHyperParam(default=False),
        'chp': CategoricalHyperParam(choices=['a', 'b', None], default=None),
        'fhp': FloatHyperParam(min=0.1, max=1.0, default=0.5),
        'ihp': IntHyperParam(min=-1, max=1)
    }
    tunable = Tunable(hyperparams)
    tuner = GPTuner(tunable)

    for _ in range(10):
        proposed = tuner.propose(1)
        tuner.record(proposed, random.random())

    # asserts
    assert len(tuner.trials) == 10
    assert len(tuner._trials_set) == 10
    assert len(tuner.raw_scores) == 10
    assert len(tuner.scores) == 10
    assert all(tuner.raw_scores == tuner.scores)


def test_tuning_minimize():
    hyperparams = {
        'bhp': BooleanHyperParam(default=False),
        'chp': CategoricalHyperParam(choices=['a', 'b', None], default=None),
        'fhp': FloatHyperParam(min=0.1, max=1.0, default=0.5),
        'ihp': IntHyperParam(min=-1, max=1)
    }
    tunable = Tunable(hyperparams)
    tuner = GPTuner(tunable, maximize=False)

    for _ in range(10):
        proposed = tuner.propose(1)
        tuner.record(proposed, random.random())

    # asserts
    assert len(tuner.trials) == 10
    assert len(tuner._trials_set) == 10
    assert len(tuner.raw_scores) == 10
    assert len(tuner.scores) == 10
    assert all(-tuner.raw_scores == tuner.scores)



# -*- coding: utf-8 -*-

"""Top level where all the acquisition functions are imported."""

from btb.tuning.acquisition.expected_improvement import ExpectedImprovementAcquisition
from btb.tuning.acquisition.predicted_score import PredictedScoreAcquisition

__all__ = ('ExpectedImprovementAcquisition', 'PredictedScoreAcquisition')



from unittest import TestCase
from unittest.mock import patch

from btb.selection.ucb1 import UCB1


class TestUCB1(TestCase):

    # METHOD: bandit(self, choice_rewards)
    # VALIDATE:
    #     * returned values
    # NOTES:
    #     * random.choice will need to be mocked

    @patch('btb.selection.ucb1.UCB1._shuffle')
    def test_bandit(self, mock__shuffle):
        """Only the choices with the highest scores are returned."""
        choices = ['DT', 'SVM', 'RF']
        mock__shuffle.return_value = choices

        # Set-up
        selector = UCB1(choices)

        # Run
        choice_rewards = {
            'DT': [0.7, 0.8, 0.9],
            'RF': [0.9, 0.93, 0.95],
            'SVM': [0.9, 0.93, 0.95]
        }

        best = selector.bandit(choice_rewards)

        # Assert
        # The first choice tried wins if there are duplicate max scores
        assert best == 'SVM'

        call_rewards = {
            'DT': [0.7, 0.8, 0.9],
            'RF': [0.9, 0.93, 0.95],
            'SVM': [0.9, 0.93, 0.95]
        }
        mock__shuffle.assert_called_once_with(call_rewards)



from unittest import TestCase

from btb.selection.selector import Selector


class TestSelector(TestCase):

    # METHOD: __init__(self, choices, **kwargs)
    # VALIDATE:
    #     * attribute value

    def test___init__(self):

        # Run
        selector = Selector(['RF', 'SVM'])

        # Assert
        assert selector.choices == ['RF', 'SVM']

    # METHOD: compute_rewards(self, scores)
    # VALIDATE:
    #     * returned value
    def test_compute_rewards(self):

        # Set-up
        selector = Selector(['RF', 'SVM'])

        # Run
        scores = [0.8, 0.9]
        rewards = selector.compute_rewards(scores)

        # Assert
        assert rewards == scores

    # METHOD: bandit(self, choice_rewards)
    # VALIDATE:
    #     * returned value

    def test_bandit(self):

        # Set-up
        selector = Selector(['RF', 'SVM'])

        # Run
        choice_rewards = {
            'RF': [0.8, 0.83, 0.85],
            'SVM': [0.9, 0.93, 0.95]
        }
        best = selector.bandit(choice_rewards)

        # Assert
        assert best == 'SVM'

    # METHOD: select(self, choice_scores)
    # VALIDATE:
    #     * returned values
    # TODO:
    #     * "if not condition: continue" should rather be "if condition: do something"

    def test_select(self):

        # Set-up
        selector = Selector(['RF', 'SVM'])

        # Run
        choice_scores = {
            'DT': [0.7, 0.73, 0.75],
            'RF': [0.8, 0.83, 0.85],
            'SVM': [0.9, 0.93, 0.95]
        }
        best = selector.select(choice_scores)

        # Assert
        assert best == 'SVM'



from unittest import TestCase
from unittest.mock import patch

from btb.selection.custom_selector import CustomSelector


class TestCustomSelector(TestCase):

    # METHOD: select(self, choice_scores)
    # VALIDATE:
    #     * returned values and randint call
    # NOTES:
    #     * randint will need to be mocked

    @patch('btb.selection.custom_selector.random')
    def test_select(self, random_mock):

        # Set-up
        selector = CustomSelector(['DT', 'RF', 'SVM'])

        random_mock.randint.return_value = 0

        # Run
        choice_scores = {
            'DT': [0.7, 0.73, 0.75],
            'RF': [0.8, 0.83, 0.85],
            'SVM': [0.9, 0.93, 0.95]
        }
        best = selector.select(choice_scores)

        # Assert
        assert best == 'DT'
        random_mock.randint.assert_called_once_with(0, 2)



# -*- coding: utf-8 -*-

"""Top level where all the hyperparameters are imported."""

from btb.tuning.hyperparams.boolean import BooleanHyperParam
from btb.tuning.hyperparams.categorical import CategoricalHyperParam
from btb.tuning.hyperparams.numerical import FloatHyperParam, IntHyperParam

__all__ = ('BooleanHyperParam', 'CategoricalHyperParam', 'FloatHyperParam', 'IntHyperParam')



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock, patch

import numpy as np

from btb.tuning.hyperparams.base import BaseHyperParam


class TestBaseHyperParam(TestCase):

    def test__within_range_valid_range(self):
        # setup
        instance = MagicMock()
        values = np.array([[1], [2]])
        _min = 0
        _max = 10

        # run / assert
        BaseHyperParam._within_range(instance, values, min=_min, max=_max)

    def test__within_range_invalid_range(self):
        # setup
        instance = MagicMock()
        values = np.array([[1], [2]])
        _min = 0
        _max = 1

        # run / assert
        with self.assertRaises(ValueError):
            BaseHyperParam._within_range(instance, values, min=_min, max=_max)

    def test__within_search_space(self):
        # setup
        instance = MagicMock()
        values = np.array(15)

        # run
        BaseHyperParam._within_search_space(instance, values)

        # assert
        instance._within_range.assert_called_once_with(np.array(15), min=0, max=1)

    def test__within_hyperparam_space(self):
        # setup
        instance = MagicMock()
        instance.min = 0
        instance.max = 10
        values = 5

        # run
        BaseHyperParam._within_hyperparam_space(instance, values)

        # assert
        instance._within_range.assert_called_once_with(5, min=0, max=10)

    def test_inverse_transform(self):
        # setup
        instance = MagicMock()
        instance._inverse_transform = MagicMock(return_value=3)
        instance._to_array.return_value = 2
        values = 1

        # run
        result = BaseHyperParam.inverse_transform(instance, values)

        # assert
        instance._to_array.assert_called_once_with(1)
        instance._within_search_space.assert_called_once_with(2)
        instance._inverse_transform.assert_called_once_with(2)

        assert result == 3

    @patch('btb.tuning.hyperparams.base.np.asarray')
    def test_transform_values_not_ndarray(self, mock_np_asarray):
        # setup
        instance = MagicMock()
        instance._transform.return_value = 2
        mock_np_asarray.return_value = np.array([[1]])
        values = 1

        # run
        result = BaseHyperParam.transform(instance, values)

        # assert
        mock_np_asarray.asser_called_once_with(1)
        instance._within_hyperparam_space.assert_called_once_with(np.array([[1]]))
        instance._transform.assert_called_once_with(np.array([[1]]))
        self.assertEqual(result, 2)

    def test_transform_values_list(self):
        # setup
        instance = MagicMock()
        instance._transform.return_value = 2
        values = np.array([[1]])

        # run
        result = BaseHyperParam.transform(instance, values)

        # assert
        instance._within_hyperparam_space.assert_called_once_with(np.array([[1]]))
        instance._transform.assert_called_once_with(np.array([[1]]))
        assert result == 2

    @patch('btb.tuning.hyperparams.base.np.asarray')
    def test_transform_dimensions_gt_two(self, mock_asarray):
        # setup
        array = MagicMock()
        array.shape.__len__.return_value = 3

        mock_asarray.return_value = array

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam.transform(MagicMock(), 1)

    def test__to_array_scalar_value_dimension_one(self):
        # setup
        instance = MagicMock()
        instance.dimensions = 1
        values = 1

        # run
        result = BaseHyperParam._to_array(instance, values)

        # assert
        np.testing.assert_array_equal(result, np.array([[1]]))

    def test__to_array_scalar_value_dimensions_gt_one(self):
        # setup
        instance = MagicMock()
        instance.dimensions = 2

        values = 1

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    def test__to_array_list_values_of_scalar_values_dimensions_one(self):
        # setup
        instance = MagicMock()
        instance.dimensions = 1
        values = [1, 2, 3]

        # run
        result = BaseHyperParam._to_array(instance, values)

        # assert
        np.testing.assert_array_equal(result, np.array([[1], [2], [3]]))

    def test__to_array_list_values_of_scalar_values_dimensions_gt_one(self):
        # setup
        instance = MagicMock()
        instance.dimensions = 2
        values = [1, 2, 3]

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    def test__to_array_list_values_of_list_values_dimensions_one(self):
        # setup
        instance = MagicMock()
        instance.dimensions = 1
        values = [[1], [2], [3]]

        # run
        result = BaseHyperParam._to_array(instance, values)

        # assert
        np.testing.assert_array_equal(result, np.array([[1], [2], [3]]))

    def test__to_array_list_values_of_list_values_dimensions_two(self):

        # setup
        instance = MagicMock()
        instance.dimensions = 2
        values = [[1], [2], [3]]

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    def test__to_array_list_values_of_list_values_one_scalar(self):
        # setup
        instance = MagicMock()
        instance.dimensions = 1
        values = [[1], 2, [3]]

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    def test__to_array_values_shape_gt_two(self):
        # setup
        instance = MagicMock()
        instance.dimensions = 1
        values = [[[1], [3]]]

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    @patch('btb.tuning.hyperparams.base.np.array')
    def test__to_array_len_shape_is_one(self, mock_np_array):
        # setup
        instance = MagicMock()
        instance.dimensions = 1
        values = [1]
        array = MagicMock()
        array.shape.__len__.return_value = 1

        mock_np_array.return_value = array

        # run
        BaseHyperParam._to_array(instance, values)

        # assert
        array.reshape.assert_called_once_with(-1, 1)

    @patch('btb.tuning.hyperparams.base.np.array')
    def test__to_array_more_than_one_column_for_dimensions_one(self, mock_np_array):
        # setup
        instance = MagicMock()
        instance.dimensions = 1
        values = [1]
        array = MagicMock()
        array.shape.__len__.return_value = 2
        array.shape = [2, 3]

        mock_np_array.return_value = array

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    @patch('btb.tuning.hyperparams.base.np.array')
    def test__to_array_values_shape_one_dimensions_two(self, mock_np_array):

        # setup
        instance = MagicMock()
        instance.dimensions = 2
        values = [1]
        array = MagicMock()
        array.__len__.return_value = 3
        array.shape.__len__.return_value = 1

        mock_np_array.return_value = array

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    @patch('btb.tuning.hyperparams.base.np.isscalar')
    @patch('btb.tuning.hyperparams.base.np.array')
    def test__to_array_values_not_scalar_dimensions_two(self, mock_np_array, mock_np_isscalar):

        # setup
        instance = MagicMock()
        instance.dimensions = 2
        mock_np_isscalar.return_value = False
        values = [1]
        array = MagicMock()
        array.__len__.return_value = 2
        array.shape.__len__.return_value = 1

        mock_np_array.return_value = array

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    @patch('btb.tuning.hyperparams.base.np.isscalar')
    @patch('btb.tuning.hyperparams.base.np.array')
    def test__to_array_values_reshape_dimensions_two(self, mock_np_array, mock_np_isscalar):

        # setup
        mock_np_isscalar.side_effect = [False, True]

        instance = MagicMock()
        instance.dimensions = 2
        mock_np_isscalar.return_value = False
        values = [1]
        array = MagicMock()
        array.__len__.return_value = 2
        array.shape.__len__.return_value = 1

        mock_np_array.return_value = array

        # run
        BaseHyperParam._to_array(instance, values)
        array.reshape.assert_called_once_with(1, -1)

    @patch('btb.tuning.hyperparams.base.np.array')
    def test__to_array_len_shape_is_gt_two(self, mock_np_array):
        # setup
        instance = MagicMock()
        instance.dimensions = 1
        values = [1]
        array = MagicMock()
        array.shape.__len__.return_value = 3

        mock_np_array.return_value = array

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)

    @patch('btb.tuning.hyperparams.base.np.isscalar')
    @patch('btb.tuning.hyperparams.base.np.array')
    def test__to_array_not_all_scalars(self, mock_np_array, mock_np_isscalar):
        # setup
        instance = MagicMock()
        instance.dimensions = 2
        values = [1, 2]
        array = MagicMock()
        array.shape.__len__.return_value = 1
        array.__len__.return_value = 2
        array.__iter__.return_value = [1]

        mock_np_isscalar.side_effect = [False, True, True, False, False]
        mock_np_array.return_value = array

        # run
        with self.assertRaises(ValueError):
            BaseHyperParam._to_array(instance, values)



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock, call

import numpy as np

from btb.tuning.acquisition.predicted_score import PredictedScoreAcquisition


def assert_called_with_np_array(mock_calls, real_calls):
    assert len(mock_calls) == len(real_calls)

    for mock_call, real_call in zip(mock_calls, real_calls):
        np.testing.assert_array_equal(mock_call[0][0], real_call[1][0])
        assert mock_call[0][1] == real_call[1][1]


class TestPredictedScoreAcquisition(TestCase):

    def test__acquire(self):
        # setup
        candidates = np.array([[1, 2]])
        instance = MagicMock()
        instance._get_max_candidates.return_value = 'max_candidate'

        # assert
        result = PredictedScoreAcquisition._acquire(instance, candidates)

        # assert
        called_with = instance._get_max_candidates.call_args_list
        expected_call = [call(np.array([1]), 1)]
        assert_called_with_np_array(called_with, expected_call)
        assert result == 'max_candidate'

    def test__acquire_num_candidates_gt_one(self):
        # setup
        candidates = np.array([[1, 1], [2, 2], [3, 3], [4, 4]])
        instance = MagicMock()
        instance._get_max_candidates.return_value = 'max_candidates'

        # assert
        result = PredictedScoreAcquisition._acquire(instance, candidates, num_candidates=2)

        # assert
        called_with = instance._get_max_candidates.call_args_list
        expected_call = [call(np.array([1, 2, 3, 4]), 2)]
        assert_called_with_np_array(called_with, expected_call)
        assert result == 'max_candidates'

    def test__acquire_candidates_shape_one(self):
        # setup
        candidates = np.array([1, 2, 3])
        instance = MagicMock()
        instance._get_max_candidates.return_value = 'max_candidate'

        # assert
        result = PredictedScoreAcquisition._acquire(instance, candidates, num_candidates=1)

        # assert
        called_with = instance._get_max_candidates.call_args_list
        expected_call = [call(np.array([1, 2, 3]), 1)]
        assert_called_with_np_array(called_with, expected_call)
        assert result == 'max_candidate'



# -*- coding: utf-8 -*-

"""Package where the tuners based on GaussianProcessMetaModel are defined."""

from btb.tuning.acquisition.expected_improvement import ExpectedImprovementAcquisition
from btb.tuning.acquisition.predicted_score import PredictedScoreAcquisition
from btb.tuning.metamodels.gaussian_process import GaussianProcessMetaModel
from btb.tuning.tuners.base import BaseMetaModelTuner


class GPTuner(GaussianProcessMetaModel, PredictedScoreAcquisition, BaseMetaModelTuner):
    """Gaussian Process Tuner.

    This class uses a ``GaussianProcessRegressor`` model from the ``sklearn.gaussian_process``
    package, using a ``numpy.argmax`` function to return the better configurations predicted
    from the model.
    """
    def __init__(self, tunable, maximize=True, num_candidates=1000,
                 min_trials=2, length_scale=0.1):
        """Create an instance of ``GPTuner``.

        Args:
            tunable (btb.tuning.tunable.Tunable):
                Instance of a tunable class containing hyperparameters to be tuned.
            num_candidates (int):
                Number of samples to generate and select the best of it for each proposal.
                Defaults to 1000.
            maximize (bool):
                If ``True`` the model will understand that the score bigger is better, if ``False``
                the smaller is better. Defaults to ``True``.
            min_trials (int):
                Number of recorded ``trials`` needed to perform a fitting over the model.
                Defaults to 2.
            length_scale (float or array):
                A float or array with shape ``(n_features,)``, used for the default ``RBF`` kernel.
        """
        self._metamodel_kwargs = {'length_scale': length_scale}
        super().__init__(tunable, maximize, num_candidates, min_trials)

    def __repr__(self):
        length_scale = self._metamodel_kwargs.get('length_scale')
        args = (self.tunable, self.maximize, self.num_candidates, self.min_trials, length_scale)
        return ('GPTuner(tunable={}, maximize={},'
                'num_candidates={}, min_trials={},'
                'length_scale={})').format(*args)


class GPEiTuner(GaussianProcessMetaModel, ExpectedImprovementAcquisition, BaseMetaModelTuner):
    """Gaussian Process Expected Improvement Tuner.

    This class uses a ``GaussianProcessRegressor`` model from the ``sklearn.gaussian_process``
    package, using an ``ExpectedImprovement`` function to return the better configurations
    predicted from the model.
    """
    def __init__(self, tunable, maximize=True, num_candidates=1000,
                 min_trials=2, length_scale=0.1):
        """Create an instance of ``GPEiTuner``.

        Args:
            tunable (btb.tuning.tunable.Tunable):
                Instance of a tunable class containing hyperparameters to be tuned.
            num_candidates (int):
                Number of samples to generate and select the best of it for each proposal.
                Defaults to 1000.
            maximize (bool):
                If ``True`` the model will understand that the score bigger is better, if ``False``
                the smaller is better. Defaults to ``True``.
            min_trials (int):
                Number of recorded ``trials`` needed to perform a fitting over the model.
                Defaults to 2.
            length_scale (float or array):
                A float or array with shape ``(n_features,)``, used for the default ``RBF`` kernel.
        """
        self.length_scale = length_scale
        self._metamodel_kwargs = {'length_scale': self.length_scale}
        super().__init__(tunable, maximize, num_candidates, min_trials)

    def __repr__(self):
        length_scale = self._metamodel_kwargs.get('length_scale')
        args = (self.tunable, self.maximize, self.num_candidates, self.min_trials, length_scale)
        return ('GPTuner(tunable={}, maximize={},'
                'num_candidates={}, min_trials={},'
                'length_scale={})').format(*args)



# -*- coding: utf-8 -*-

"""Top-level package for BTB."""

from btb.session import BTBSession

__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '0.3.8'
__all__ = ('BTBSession', )



from unittest import TestCase
from unittest.mock import patch

from btb.selection.uniform import Uniform


class TestUniform(TestCase):

    # METHOD: select(self, choice_scores)
    # VALIDATE:
    #     * returned values and randint call
    # NOTES:
    #     * randint will need to be mocked

    @patch('random.choice')
    def test_select(self, mock_choice):

        # Set-up
        choices = ['DT', 'RF', 'SVM']
        selector = Uniform(choices)
        expected = 'DT'

        mock_choice.return_value = expected

        # Run
        choice_scores = None
        actual = selector.select(choice_scores)

        # Assert
        assert actual == expected
        mock_choice.assert_called_once_with(choices)



# -*- coding: utf-8 -*-

from unittest import TestCase

import pytest

from btb.session import BTBSession
from btb.tuning import StopTuning


class BTBSessionTest(TestCase):

    @staticmethod
    def scorer(name, proposal):
        """score = name length + parameter.

        best proposal will be `a_tunable + a_parameter=0`
        """
        return len(name) + proposal['a_parameter']

    def test_stop(self):
        tunables = {
            'a_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 2]
                }
            }
        }

        session = BTBSession(tunables, self.scorer)

        with pytest.raises(StopTuning):
            session.run()

    def test_maximize(self):
        tunables = {
            'a_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 2]
                }
            }
        }

        session = BTBSession(tunables, self.scorer)

        best = session.run(3)

        assert best == session.best_proposal

        assert best['name'] == 'a_tunable'
        assert best['config'] == {'a_parameter': 2}

    def test_minimize(self):
        tunables = {
            'a_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 2]
                }
            }
        }

        session = BTBSession(tunables, self.scorer, maximize=False)

        best = session.run(3)

        assert best == session.best_proposal
        assert best['name'] == 'a_tunable'
        assert best['config'] == {'a_parameter': 0}

    def test_multiple(self):
        tunables = {
            'a_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 2]
                }
            },
            'another_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 2]
                }
            }
        }

        session = BTBSession(tunables, self.scorer)

        best = session.run(6)

        assert best['name'] == 'another_tunable'
        assert best['config'] == {'a_parameter': 2}

    def test_errors(self):
        tunables = {
            'a_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 2]
                }
            },
            'another_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 2]
                }
            }
        }

        def scorer(name, proposal):
            if name == 'another_tunable':
                raise Exception()
            else:
                return proposal['a_parameter']

        session = BTBSession(tunables, scorer)

        best = session.run(4)

        assert best['name'] == 'a_tunable'
        assert best['config'] == {'a_parameter': 2}

    def test_normalized_score_becomes_none(self):
        """Tunables that worked at some point but end up removed are not tried again.

        After commit ``6a08dc3cf1b68b35630cae6a87783aec4e2c9f83`` the following
        scenario has been observed:

        - One tunable produces a score at least once and then fails the next trials.
        - All the other tunables never produce any score.
        - Once all the tuners are created, only the one that produced a score is used.
        - After enough errors, this one is discarded, so `_normalized_errors` is empty.
        - Since a random.choice is used over the list of tunables, which still contains
          the one tha has been discarded, at some point the discarded one is tried again.

        This test certifies that this scenario cannot happen again, by validating that
        the number of errors is always ``max_errors`` at most.
        """
        scores = []

        def scorer(name, proposal):
            """Produce a score for the first trial and then fail forever."""
            if not scores:
                scores.append(1)   # boolean variable fails due to scope unles using global
                return 1

            raise Exception()

        tunables = {
            'a_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 10]
                }
            },
            'another_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 10]
                }
            }
        }

        session = BTBSession(tunables, scorer, max_errors=3)

        with pytest.raises(StopTuning):
            session.run(8)

        assert session.errors == {'a_tunable': 3, 'another_tunable': 3}

    @pytest.mark.skip(reason="This is not implemented yet")
    def test_allow_duplicates(self):
        tunables = {
            'a_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 2]
                }
            }
        }

        session = BTBSession(tunables, self.scorer, allow_duplicates=True)

        best = session.run(10)

        assert best['name'] == 'another_tunable'
        assert best['config'] == {'a_parameter': 2}

    def test_allow_errors(self):
        tunables = {
            'a_tunable': {
                'a_parameter': {
                    'type': 'int',
                    'default': 0,
                    'range': [0, 1]
                }
            }
        }

        def scorer(name, proposal):
            if proposal['a_parameter'] == 0:
                raise Exception()

            return 1

        session = BTBSession(tunables, scorer, max_errors=10)

        best = session.run(10)

        assert best['name'] == 'a_tunable'
        assert best['config'] == {'a_parameter': 1}



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock

from btb.tuning.tunable import Tunable
from btb.tuning.tuners.uniform import UniformTuner


class TestUniformTuner(TestCase):
    """Test UniformTuner class."""

    def setUp(self):
        tunable = MagicMock(spec_set=Tunable)
        self.instance = UniformTuner(tunable)

    def test___init__(self):
        assert isinstance(self.instance.tunable, MagicMock)

    def test__propose_no_duplicates(self):
        """Test that `_propose` method calls it's parent `_sample` method with a single value
        and `allow_duplicates=False`.
        """
        # setup
        self.instance._sample = MagicMock(return_value=[1])

        # run
        result = self.instance._propose(1, False)

        # assert
        assert result == [1]
        self.instance._sample.assert_called_once_with(1, False)

    def test__propose_allow_duplicates(self):
        """Test that `_propose` method calls it's parent `_sample` method with a single value
        and `allow_duplicates=True`.
        """
        # setup
        self.instance._sample = MagicMock(return_value=[1])

        # run
        result = self.instance._propose(1, True)

        # assert
        assert result == [1]
        self.instance._sample.assert_called_once_with(1, True)



# -*- coding: utf-8 -*-

from unittest import TestCase

import numpy as np

from btb.tuning.acquisition.expected_improvement import ExpectedImprovementAcquisition


class TestExpectedImprovementAcquisition(TestCase):

    def test__acquire(self):
        # run
        instance = ExpectedImprovementAcquisition()
        instance.scores = np.array([0.5, 0.6, 0.7])

        predictions = np.array([
            [0.8, 1],
            [0.9, 2]
        ])
        best = instance._acquire(predictions)

        # assert
        np.testing.assert_array_equal(best, np.array([1]))

    def test__acquire_n_candidates(self):
        # run
        instance = ExpectedImprovementAcquisition()
        instance.scores = np.array([0.5, 0.9, 0.7, 0.8])

        predictions = np.array([
            [0.1, 1],
            [0.9, 9],
            [0.7, 3],
            [0.75, 4]
        ])
        best = instance._acquire(predictions, 2)

        # assert
        np.testing.assert_array_equal(best, np.array([1, 3]))



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import patch

import numpy as np

from btb.tuning.hyperparams.boolean import BooleanHyperParam


class TestBooleanHyperParam(TestCase):
    """Unit test for the class ``BooleanHyperParam``."""

    @classmethod
    def setUpClass(cls):
        """Instantiate a BooleanHyperParam to be used in all the tests."""
        cls.instance = BooleanHyperParam()

    def test__within_hyperparam_space_values_in_space(self):
        """Test that the method ``_within_hyperparam_space`` does not raise a value error when
        values in the hyperparameter space are given."""
        # setup
        values = np.array([[True], [False]])

        # run assert
        self.instance._within_hyperparam_space(values)

    def test__within_hyperparam_space_values_out_of_space(self):
        """Test that the method ``_within_hyperparam_space`` does not raise a value error when
        values in the hyperparameter space are given."""
        # setup
        values = np.array([[1], [0]])

        # run/assert
        with self.assertRaises(ValueError):
            self.instance._within_hyperparam_space(values)

    def test__inverse_transform_single_value(self):
        """Test the method ``_inverse_transform`` to perform a denomalization over a value from
        the search space [0, 1]^K.
        """
        # setup
        values = np.array([[1]])

        # run
        result = self.instance._inverse_transform(values)

        # assert
        np.testing.assert_array_equal(result, np.array([[True]]))

    def test__inverse_transform_multiple_values(self):
        """Test the method ``_inverse_transform`` to perform a denomalization over multiple values
        from the search space [0, 1]^K.
        """
        # setup
        values = np.array([[1], [0], [0], [1]])

        # run
        result = self.instance._inverse_transform(values)

        # assert
        expected_result = np.array([[True], [False], [False], [True]])

        np.testing.assert_array_equal(result, expected_result)

    def test__transform_single_value(self):
        """Test that the method ``_transform`` performs a normalization over a ``boolean`` value
        and converts it in to a search space of [0, 1]^k.
        """

        # setup
        values = np.array([[True]])

        # run
        result = self.instance._transform(values)

        # assert
        np.testing.assert_array_equal(result, np.array([[1]]))

    def test__transform_multiple_values(self):
        """Test that the method ``_transform`` performs a normalization over ``boolean`` values and
        converts them in to a search space of [0, 1]^k.
        """

        # setup
        values = np.array([[True], [False]])

        # run
        result = self.instance._transform(values)

        # assert
        np.testing.assert_array_equal(result, np.array([[1], [0]]))

    @patch('btb.tuning.hyperparams.boolean.np.random.random')
    def test_sample(self, mock_np_random):
        """Test that the method ``sample`` returns values from the search space and not the
        original hyperparameter space.
        """
        # setup
        n = 4
        mock_np_random.return_value = np.array([[0.1], [0.2], [0.5], [0.99]])

        # run
        result = self.instance.sample(n)

        # assert
        expected_result = np.array([[0], [0], [0], [1]])

        mock_np_random.assert_called_once_with((4, 1))
        self.assertEqual(len(result), 4)
        np.testing.assert_array_equal(result, expected_result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup


try:
    with open('README.md') as readme_file:
        readme = readme_file.read()

except IOError:
    readme = ''


try:
    with open('HISTORY.md') as history_file:
        history = history_file.read()

except IOError:
    history = ''


install_requires = [
    'numpy>=1.14.0,<1.18.0',
    'scikit-learn>0.20.0,<0.23',
    'scipy>=1.0.1,<1.4.0',
    'pandas>=0.21.0,<0.26.0',
    'tqdm>=4.36.1,<4.50.0',
]


tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
]


setup_requires = [
    'pytest-runner>=2.11.1',
]


development_requires = [
    # general
    'pip>=9.0.1',
    'bumpversion>=0.5.3,<0.6',
    'watchdog>=0.8.3,<0.11',

    # docs
    'm2r>=0.2.0,<0.3',
    'nbsphinx>=0.5.0,<0.7',
    'Sphinx>=1.7.1,<3',
    'sphinx_rtd_theme>=0.2.4,<0.5',

    # style check
    'flake8>=3.7.7,<4',
    'isort>=4.3.4,<5',

    # fix style issues
    'autoflake>=1.1,<2',
    'autopep8>=1.4.3,<2',

    # distribute on PyPI
    'twine>=1.10.0,<4',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1,<6',
    'tox>=2.9.1,<4',

    # others
    'jupyter>=1.0.0',
]


setup(
    author='MIT Data To AI Lab',
    author_email='dailabmit@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description='Bayesian Tuning and Bandits',
    extras_require={
        'test': tests_require,
        'dev': development_requires + tests_require,
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords='machine learning hyperparameters tuning classification',
    license='MIT license',
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    name='baytune',
    packages=find_packages(include=['btb', 'btb.*']),
    python_requires='>=3.5',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/HDI-Project/BTB',
    version='0.3.8',
    zip_safe=False,
)



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock, call, patch

import numpy as np
import pandas as pd

from btb.tuning.hyperparams.boolean import BooleanHyperParam
from btb.tuning.hyperparams.categorical import CategoricalHyperParam
from btb.tuning.hyperparams.numerical import IntHyperParam
from btb.tuning.tunable import Tunable


def assert_called_with_np_array(mock_calls, real_calls):
    assert len(mock_calls) == len(real_calls)

    for mock_call, real_call in zip(mock_calls, real_calls):
        np.testing.assert_array_equal(mock_call[0], real_call[1])


class TestTunable(TestCase):
    """Unit test for the class ``Tunable``."""

    @patch('btb.tuning.tunable.list')
    def setUp(self, list_mock):
        """Instantiate the ``Tunable`` and it's ``Hyperparameters`` that we will be using."""
        self.bhp = MagicMock(spec_set=BooleanHyperParam)
        self.chp = MagicMock(spec_set=CategoricalHyperParam)
        self.ihp = MagicMock(spec_set=IntHyperParam)

        list_mock.return_value = ['bhp', 'chp', 'ihp']

        hyperparams = {
            'bhp': self.bhp,
            'chp': self.chp,
            'ihp': self.ihp,
        }

        self.instance = Tunable(hyperparams)

    def test___init__with_given_names(self):
        """Test that the names are being generated correctly."""
        # assert
        assert self.instance.names == ['bhp', 'chp', 'ihp']

    def test_transform_valid_dict(self):
        """Test transform method with a dictionary that has all the hyperparameters."""
        # setup
        self.bhp.transform.return_value = [[1]]
        self.chp.transform.return_value = [[0]]
        self.ihp.transform.return_value = [[1]]

        values_dict = {
            'bhp': True,
            'chp': 'cat',
            'ihp': 1
        }

        # run
        result = self.instance.transform(values_dict)

        # assert
        self.bhp.transform.assert_called_once_with(True)
        self.chp.transform.assert_called_once_with('cat')
        self.ihp.transform.assert_called_once_with(1)

        np.testing.assert_array_equal(result, np.array([[1, 0, 1]]))

    def test_transform_empty_dict(self):
        """Test transform method with a dictionary that has a missing hyperparameters."""
        # run / assert
        with self.assertRaises(KeyError):
            self.instance.transform({})

    def test_transform_invalid_dict_one_missing(self):
        """Test transform method with a dictionary that has a missing hyperparameters."""
        # run / assert
        values = {
            'bhp': True,
            'chp': 'cat'
        }

        with self.assertRaises(KeyError):
            self.instance.transform(values)

    def test_transform_list_of_dicts(self):
        """Test transform method with a list of dictionaries."""
        # setup
        self.bhp.transform.return_value = [[1], [0]]
        self.chp.transform.return_value = [[0], [1]]
        self.ihp.transform.return_value = [[1], [1]]

        values_list_dict = [
            {'bhp': True, 'chp': 'cat', 'ihp': 2},
            {'bhp': False, 'chp': 'cat', 'ihp': 3}
        ]

        # run
        results = self.instance.transform(values_list_dict)

        # assert
        assert_called_with_np_array(self.bhp.transform.call_args_list, [call([True, False])])
        assert_called_with_np_array(self.chp.transform.call_args_list, [call(['cat', 'cat'])])
        assert_called_with_np_array(self.ihp.transform.call_args_list, [call([2, 3])])

        np.testing.assert_array_equal(results, np.array([[1, 0, 1], [0, 1, 1]]))

    def test_transform_list_of_invalid_dicts(self):
        """Test transform method with a list of dictionaries where one of them does not have
        the categorical value."""

        # setup
        self.bhp.transform.return_value = [[1], [0]]

        # Here we create a CHP so we can raise an value error as there will be a NaN inside the
        # pandas.DataFrame.
        self.chp = CategoricalHyperParam(['cat', 'dog'])
        self.ihp.transform.return_value = [[1], [1]]

        values_list_dict = [
            {'bhp': True, 'ihp': 2},
            {'bhp': False, 'chp': 'cat', 'ihp': 3}
        ]

        # run / assert
        with self.assertRaises(ValueError):
            self.instance.transform(values_list_dict)

    def test_transform_empty_list(self):
        """Test transform method with an empty list."""
        # run / assert
        with self.assertRaises(IndexError):
            self.instance.transform(list())

    def test_transform_valid_pandas_series(self):
        """Test transform method over a valid ``pandas.Series`` object."""
        # setup
        self.bhp.transform.return_value = [[1]]
        self.chp.transform.return_value = [[0]]
        self.ihp.transform.return_value = [[1]]

        values = pd.Series([False, 'cat', 1], index=['bhp', 'chp', 'ihp'])

        # run
        result = self.instance.transform(values)

        # assert
        self.bhp.transform.assert_called_once_with(False)
        self.chp.transform.assert_called_once_with('cat')
        self.ihp.transform.assert_called_once_with(1)

        np.testing.assert_array_equal(result, np.array([[1, 0, 1]]))

    def test_transform_invalid_pandas_series(self):
        """Test transform method over a ``pandas.Series`` object that does not have index."""
        # setup
        values = pd.Series([False, 'cat', 1])

        # run
        with self.assertRaises(KeyError):
            self.instance.transform(values)

    def test_transform_array_like_list(self):
        """Test transform a valid array like list."""
        # setup
        self.bhp.transform.return_value = [[1]]
        self.chp.transform.return_value = [[0]]
        self.ihp.transform.return_value = [[1]]

        values = [[True, 'dog', 2], [False, 'cat', 3]]

        # run
        result = self.instance.transform(values)

        # assert
        assert_called_with_np_array(
            self.bhp.transform.call_args_list,
            [call(np.array([True, False]))]
        )
        assert_called_with_np_array(
            self.chp.transform.call_args_list,
            [call(np.array(['dog', 'cat']))]
        )
        assert_called_with_np_array(
            self.ihp.transform.call_args_list,
            [call(np.array([2, 3]))]
        )

        np.testing.assert_array_equal(result, np.array([[1, 0, 1]]))

    def test_transform_simple_list(self):
        """Test that the method transform performs a transformation over a list with a single
        combination of hyperparameter valid values.
        """
        # setup
        self.bhp.transform.return_value = [[1]]
        self.chp.transform.return_value = [[0]]
        self.ihp.transform.return_value = [[1]]

        values = [True, 'dog', 2]

        # run
        result = self.instance.transform(values)

        # assert
        self.bhp.transform.assert_called_once_with(True)
        self.chp.transform.assert_called_once_with('dog')
        self.ihp.transform.assert_called_once_with(2)

        np.testing.assert_array_equal(result, np.array([[1, 0, 1]]))

    def test_transform_pd_df(self):
        """Test that the method transform performs a transformation over a ``pandas.DataFrame``
        with a single combination of hyperparameter valid values.
        """
        # setup
        self.bhp.transform.return_value = [[1]]
        self.chp.transform.return_value = [[0]]
        self.ihp.transform.return_value = [[1]]

        values = pd.DataFrame([[True, 'dog', 2]], columns=['bhp', 'chp', 'ihp'])

        # run
        result = self.instance.transform(values)

        # assert
        self.bhp.transform.assert_called_once_with(True)
        self.chp.transform.assert_called_once_with('dog')
        self.ihp.transform.assert_called_once_with(2)

        np.testing.assert_array_equal(result, np.array([[1, 0, 1]]))

    def test_transform_simple_invalid_list(self):
        """Test that the method transform does not transform a list with a single combination
        of invalid hyperparameter values.
        """
        # run / assert
        with self.assertRaises(TypeError):
            self.instance.transform([[True], 1, 2])

    def test_inverse_transform_valid_data(self):
        """Test that the inverse transform method is calling the hyperparameters."""
        # setup
        self.bhp.inverse_transform.return_value = [[True]]
        self.chp.inverse_transform.return_value = [['cat']]
        self.ihp.inverse_transform.return_value = [[1]]

        values = [[1, 0, 1]]

        # run
        result = self.instance.inverse_transform(values)

        # assert
        expected_result = pd.DataFrame(
            {
                'bhp': [True],
                'chp': ['cat'],
                'ihp': [1]
            },
            dtype=object
        )

        self.bhp.inverse_transform.assert_called_once_with([1])
        self.chp.inverse_transform.assert_called_once_with([0])
        self.ihp.inverse_transform.assert_called_once_with([1])
        pd.testing.assert_frame_equal(result, expected_result)

    def test_inverse_transform_invalid_data(self):
        """Test that the a ``TypeError`` is being raised when calling with the invalid data."""
        # setup
        values = [1, 0, 1]

        # run
        with self.assertRaises(TypeError):
            self.instance.inverse_transform(values)

    def test_sample(self):
        """Test that the method sample generates data from all the ``hyperparams``."""

        # setup
        # Values have been changed to ensure that each one of them is being called.
        self.bhp.sample.return_value = [['a']]
        self.chp.sample.return_value = [['b']]
        self.ihp.sample.return_value = [['c']]

        # run
        result = self.instance.sample(1)

        # assert
        expected_result = np.array([['a', 'b', 'c']])

        assert set(result.flat) == set(expected_result.flat)
        self.bhp.sample.assert_called_once_with(1)
        self.chp.sample.assert_called_once_with(1)
        self.ihp.sample.assert_called_once_with(1)

    def test_get_defaults(self):
        # setup
        bhp = MagicMock(default=True)
        chp = MagicMock(default='test')
        ihp = MagicMock(default=1)

        hyperparams = {
            'bhp': bhp,
            'chp': chp,
            'ihp': ihp,
        }

        self.instance = Tunable(hyperparams)

        # run
        result = self.instance.get_defaults()

        # assert
        assert result == {'bhp': True, 'chp': 'test', 'ihp': 1}

    def test_from_dict_not_a_dict(self):
        # run
        with self.assertRaises(TypeError):
            Tunable.from_dict(1)

    @patch('btb.tuning.tunable.IntHyperParam')
    @patch('btb.tuning.tunable.FloatHyperParam')
    @patch('btb.tuning.tunable.CategoricalHyperParam')
    @patch('btb.tuning.tunable.BooleanHyperParam')
    def test_from_dict(self, mock_bool, mock_cat, mock_float, mock_int):
        # setup
        mock_bool.return_value.dimensions = 1
        mock_cat.return_value .dimensions = 1
        mock_float.return_value.dimensions = 1
        mock_int.return_value.dimensions = 1

        mock_bool.return_value.cardinality = 1
        mock_cat.return_value .cardinality = 1
        mock_float.return_value.cardinality = 1
        mock_int.return_value.cardinality = 1

        # run
        hyperparameters = {
            'bhp': {
                'type': 'bool',
                'default': False
            },
            'chp': {
                'type': 'str',
                'default': 'cat',
                'range': ['a', 'b', 'cat']
            },
            'fhp': {
                'type': 'float',
                'default': None,
                'range': [0.1, 1.0]
            },
            'ihp': {
                'type': 'int',
                'default': 5,
                'range': [1, 10]
            }
        }

        result = Tunable.from_dict(hyperparameters)

        # assert
        mock_bool.assert_called_once_with(default=False)
        mock_cat.assert_called_once_with(choices=['a', 'b', 'cat'], default='cat')
        mock_float.assert_called_once_with(min=0.1, max=1.0, default=None)
        mock_int.assert_called_once_with(min=1, max=10, default=5)

        expected_tunable_hp = {
            'bhp': mock_bool.return_value,
            'chp': mock_cat.return_value,
            'fhp': mock_float.return_value,
            'ihp': mock_int.return_value
        }

        assert result.hyperparams == expected_tunable_hp
        assert result.dimensions == 4
        assert result.cardinality == 1



# -*- coding: utf-8 -*-

"""Package where all the available tuners are imported."""

from btb.tuning.tuners.gaussian_process import GPEiTuner, GPTuner
from btb.tuning.tuners.uniform import UniformTuner

__all__ = ('GPEiTuner', 'GPTuner', 'UniformTuner', )



# -*- coding: utf-8 -*-

import sys
from unittest import TestCase
from unittest.mock import patch

import numpy as np

from btb.tuning.hyperparams.numerical import FloatHyperParam, IntHyperParam


class TestFloatHyperParam(TestCase):

    def test___init__no_min_no_max(self):
        """Test instantiation with ``min=None`` and ``max=None``"""
        # run
        instance = FloatHyperParam()

        # assert
        self.assertEqual(instance.min, sys.float_info.min)
        self.assertEqual(instance.max, sys.float_info.max)
        self.assertEqual(instance.default, sys.float_info.min)

    def test___init__with_np_inf(self):
        """Test instantiation with ``min=-np.inf`` and ``max=np.inf``"""
        # run
        instance = FloatHyperParam(min=-np.inf, max=np.inf)

        # assert
        self.assertEqual(instance.min, sys.float_info.min)
        self.assertEqual(instance.max, sys.float_info.max)
        self.assertEqual(instance.default, sys.float_info.min)

    def test___init__min_no_max(self):
        """Test instantiation with ``min=n`` and ``max=None``"""
        # setup
        _min = 0.1

        # run
        instance = FloatHyperParam(min=_min, max=None)

        # assert
        self.assertEqual(instance.min, _min)
        self.assertEqual(instance.max, sys.float_info.max)
        self.assertEqual(instance.default, 0.1)

    def test___init__no_min_max(self):
        """Test instantiation with ``min=None`` and ``max=n``"""
        # setup
        _max = 0.1

        # run
        instance = FloatHyperParam(min=None, max=_max)

        # assert
        self.assertEqual(instance.min, sys.float_info.min)
        self.assertEqual(instance.max, _max)
        self.assertEqual(instance.default, sys.float_info.min)

    def test___init__min_eq_max(self):
        """Test instantiation with ``min=n`` and ``max=n``"""
        # setup
        n = 0.1

        # run / assert
        with self.assertRaises(ValueError):
            FloatHyperParam(min=n, max=n)

    def test___init__min_gt_max(self):
        """Test instantiation with ``min`` being reater than ``max``"""

        # run / assert
        with self.assertRaises(ValueError):
            FloatHyperParam(min=1, max=0)

    def test___init__min_max(self):
        """Test instantiation with ``min=n`` and ``max=x``"""
        # setup
        _min = 0.1
        _max = 0.9

        # run
        instance = FloatHyperParam(min=_min, max=_max, default=1)

        # assert
        self.assertEqual(instance.min, _min)
        self.assertEqual(instance.max, _max)
        self.assertEqual(instance.default, 1.0)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__transform_no_min_no_max(self, mock_sys):
        """Test that the method ``_transform`` performs a normalization of values between ``min``
        and ``max`` with no limit set on them.
        """
        # setup
        mock_sys.float_info.max = 1000.0  # This values can be different in each OS.
        mock_sys.float_info.min = 0.0

        instance = FloatHyperParam()
        values = np.array([[0.9], [100.8]])

        # run
        result = instance._transform(values)

        # assert
        expected_result = np.array([[0.0009],
                                    [0.1008]])

        np.testing.assert_array_equal(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__transform_min_no_max(self, mock_sys):
        """Test that the method ``_transform`` performs a normalization of values between ``min``
        and ``max`` with a min value set in.
        """
        # setup
        mock_sys.float_info.max = 10.0
        _min = 1.0

        instance = FloatHyperParam(min=_min)
        values = np.array([[7.3], [4.6]])

        # run
        result = instance._transform(values)

        # assert
        expected_result = np.array([[0.7],
                                    [0.4]])

        np.testing.assert_allclose(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__transform_no_min_max(self, mock_sys):
        """Test that the method ``_transform`` performs a normalization of values between ``min``
        and ``max`` with a max value set in.
        """
        # setup
        mock_sys.float_info.min = 0.0
        _max = 1.0

        instance = FloatHyperParam(max=_max)
        values = np.array([[0.1], [0.9]])

        # run
        result = instance._transform(values)

        # assert
        expected_result = np.array([[0.1],
                                    [0.9]])

        np.testing.assert_array_equal(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__transform_min_max(self, mock_sys):
        """Test that the method ``_transform`` performs a normalization of values between ``min``
        and ``max`` with a max and min value set in.
        """
        # setup
        _min = 0.0
        _max = 10.0
        instance = FloatHyperParam(min=_min, max=_max)
        values = np.array([[0.1], [0.9]])

        # run
        result = instance._transform(values)

        # assert
        expected_result = np.array([[0.01],
                                    [0.09]])

        np.testing.assert_array_equal(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__inverse_transform_no_min_no_max(self, mock_sys):
        """Test that the method ``_inverse_transform`` performs a normalization of values between
        ``min`` and ``max`` with no min or max set.
        """
        # setup
        mock_sys.float_info.max = 1000.0
        mock_sys.float_info.min = 0.0

        instance = FloatHyperParam()
        values = np.array([[0.1], [0.2]])

        # run
        result = instance._inverse_transform(values)

        # assert
        expected_result = np.array([[100.],
                                    [200.]])

        np.testing.assert_array_equal(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__inverse_transform_min_no_max(self, mock_sys):
        """Test that the method ``_inverse_transform`` performs a normalization of values between
        ``min`` and ``max`` with min value set up.
        """
        # setup
        mock_sys.float_info.max = 10.0
        _min = 1.0

        instance = FloatHyperParam(min=_min)
        values = np.array([[0.1], [0.]])

        # run
        result = instance._inverse_transform(values)

        # assert
        expected_result = np.array([[1.9],
                                    [1.0]])

        np.testing.assert_array_equal(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__inverse_transform_no_min_max(self, mock_sys):
        """Test that the method ``_inverse_transform`` performs a normalization of values between
        ``min`` and ``max`` with a max value set up.
        """
        # setup
        mock_sys.float_info.min = 0.0
        _max = 1.0

        instance = FloatHyperParam(max=_max)
        values = np.array([[0.1], [0.9]])

        # run
        result = instance._inverse_transform(values)

        # assert
        expected_result = np.array([[0.1],
                                    [0.9]])

        np.testing.assert_array_equal(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__inverse_transform_min_max(self, mock_sys):
        """Test that the method ``_inverse_transform`` performs a normalization of values between
        ``min`` and ``max`` with a min and max value set up.
        """
        # setup
        _min = 0.0
        _max = 10.0

        instance = FloatHyperParam(min=_min, max=_max)
        values = np.array([[0.1], [0.9]])

        # run
        result = instance._inverse_transform(values)

        # assert
        expected_result = np.array([[1.],
                                    [9.]])

        np.testing.assert_array_equal(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.np.random.random')
    def test_sample(self, mock_np_random):
        """Test that the method ``sample`` is being called with `n_samples and
        `self.dimensions`.
        """
        # setup
        mock_np_random.return_value = np.array([[0.1], [0.2]])
        instance = FloatHyperParam()
        n_samples = 2

        # run
        result = instance.sample(n_samples)

        # assert
        expected_result = np.array([[0.1], [0.2]])

        mock_np_random.assert_called_once_with((n_samples, instance.dimensions))
        np.testing.assert_array_equal(result, expected_result)


class TestIntHyperParam(TestCase):

    def test___init__no_min_no_max(self):
        """Test instantiation with ``min=None`` and ``max=None``"""
        # run
        instance = IntHyperParam()

        # assert
        expected_min = -(sys.maxsize / 2)
        expected_max = sys.maxsize / 2

        self.assertEqual(instance.min, expected_min)
        self.assertEqual(instance.max, expected_max)
        self.assertEqual(instance.default, expected_min)
        self.assertEqual(instance.step, 1)

    def test___init__exclude_min_no_max(self):
        """Test instantiation with ``min=None`` and ``max=None`` excluding ``min``."""
        # run
        instance = IntHyperParam(include_min=False)

        # assert
        expected_min = int(-(sys.maxsize / 2)) + 1
        expected_max = int(sys.maxsize / 2)

        self.assertEqual(instance.min, expected_min)
        self.assertEqual(instance.max, expected_max)
        self.assertEqual(instance.step, 1)

    def test___init__no_min_exclude_max(self):
        """Test instantiation with ``min=None`` and ``max=None`` excluding ``max``."""
        # run
        instance = IntHyperParam(include_max=False)

        # assert
        expected_min = int(-(sys.maxsize / 2))
        expected_max = int(sys.maxsize / 2) - 1

        self.assertEqual(instance.min, expected_min)
        self.assertEqual(instance.max, expected_max)
        self.assertEqual(instance.default, expected_min)
        self.assertEqual(instance.step, 1)

    def test___init__min_no_max(self):
        """Test instantiation with ``min=n`` and ``max=None``"""
        # setup
        _min = 1

        # run
        instance = IntHyperParam(min=_min, max=None)

        # assert
        self.assertEqual(instance.min, _min)
        self.assertEqual(instance.max, sys.maxsize / 2)
        self.assertEqual(instance.default, 1)
        self.assertEqual(instance.step, 1)

    def test___init__no_min_max(self):
        """Test instantiation with ``min=None`` and ``max=n``"""
        # setup
        _max = 1

        # run
        instance = IntHyperParam(min=None, max=_max)

        # assert
        expected_min = -(sys.maxsize / 2)
        expected_max = 1
        self.assertEqual(instance.min, expected_min)
        self.assertEqual(instance.max, expected_max)
        self.assertEqual(instance.step, 1)

    def test___init__min_eq_max(self):
        """Test instantiation with ``min=n`` and ``max=n``"""
        # setup
        n = 1

        # run / assert
        with self.assertRaises(ValueError):
            IntHyperParam(min=n, max=n)

    def test___init__min_gt_max(self):
        """Test instantiation with ``min`` being reater than ``max``"""

        # run / assert
        with self.assertRaises(ValueError):
            IntHyperParam(min=1, max=0)

    def test___init__min_max(self):
        """Test instantiation with ``min=n`` and ``max=x``"""
        # setup
        _min = 1
        _max = 9

        # run
        instance = IntHyperParam(min=_min, max=_max, default=5)

        # assert
        self.assertEqual(instance.min, _min)
        self.assertEqual(instance.max, _max)
        self.assertEqual(instance.default, 5)
        self.assertEqual(instance.step, 1)

    def test___init__min_max_step(self):
        """Test instantiation with ``min=n`` and ``max=x`` and step."""
        # setup
        _min = 0
        _max = 10
        _step = 2

        # run
        instance = IntHyperParam(min=_min, max=_max, step=_step)

        # assert
        self.assertEqual(instance.min, 0)
        self.assertEqual(instance.max, 10)
        self.assertEqual(instance.step, 2)

    def test___init__min_max_invalid_step(self):
        """Test instantiation with ``min=n`` and ``max=x``"""
        # setup
        _min = 1
        _max = 9
        _step = 5

        # run / assert
        with self.assertRaises(ValueError):
            IntHyperParam(min=_min, max=_max, step=_step)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__transform_no_min_no_max(self, mock_sys):
        """Test that the method ``_transform`` performs a normalization of values between ``min``
        and ``max`` with no limit set on them.
        """
        # setup
        mock_sys.maxsize = 1000  # This values can be different in each OS.
        instance = IntHyperParam()
        values = np.array([[9], [100]])

        # run
        result = instance._transform(values)

        # assert
        expected_result = np.array([[0.50899101],
                                    [0.5999001]])

        np.testing.assert_allclose(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__transform_min_no_max(self, mock_sys):
        """Test that the method ``_transform`` performs a normalization of values between ``min``
        and ``max`` with a min value set in.
        """
        # setup
        mock_sys.maxsize = 1000
        _min = 1
        instance = IntHyperParam(min=_min)
        values = np.array([[5], [1]])

        # run
        result = instance._transform(values)

        # assert
        expected_result = np.array([[0.009],
                                    [0.001]])

        np.testing.assert_allclose(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__transform_no_min_max(self, mock_sys):
        """Test that the method ``_transform`` performs a normalization of values between ``min``
        and ``max`` with a max value set in.
        """
        # setup
        mock_sys.maxsize = 1000
        _max = 10
        instance = IntHyperParam(max=_max)
        values = np.array([[1], [9]])

        # run
        result = instance._transform(values)

        # assert
        expected_result = np.array([[0.981409],
                                    [0.99706458]])

        np.testing.assert_allclose(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__transform_min_max(self, mock_sys):
        """Test that the method ``_transform`` performs a normalization of values between ``min``
        and ``max`` with a max and min value set in.
        """
        # setup
        _min = 0
        _max = 10

        instance = IntHyperParam(min=_min, max=_max)
        values = np.array([[9], [1]])

        # run
        result = instance._transform(values)

        # assert
        expected_result = np.array([[0.86363636],
                                    [0.13636364]])

        np.testing.assert_allclose(result, expected_result)

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__inverse_transform_no_min_no_max(self, mock_sys):
        """Test that the method ``_inverse_transform`` performs a normalization of values between
        ``min`` and ``max`` with no min or max set.
        """
        # setup
        mock_sys.maxsize = 1000
        instance = IntHyperParam()
        values = np.array([[0.0009], [0.1008]])

        # run
        result = instance._inverse_transform(values)

        # assert
        expected_result = np.array([[-500],
                                    [-400]])

        np.testing.assert_array_equal(result, expected_result.astype(int))

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__inverse_transform_min_no_max(self, mock_sys):
        """Test that the method ``_inverse_transform`` performs a normalization of values between
        ``min`` and ``max`` with min value set up.
        """
        # setup
        mock_sys.maxsize = 1000
        _min = 1
        instance = IntHyperParam(min=_min)
        values = np.array([[0.1], [0.]])

        # run
        result = instance._inverse_transform(values)

        # assert
        expected_result = np.array([[50],
                                    [1]])

        np.testing.assert_array_equal(result, expected_result.astype(int))

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__inverse_transform_no_min_max(self, mock_sys):
        """Test that the method ``_inverse_transform`` performs a normalization of values between
        ``min`` and ``max`` with a max value set up.
        """
        # setup
        mock_sys.maxsize = 1000
        _max = 500
        instance = IntHyperParam(max=_max)
        values = np.array([[0.1], [0.9]])

        # run
        result = instance._inverse_transform(values)

        # assert
        expected_result = np.array([[-400],
                                    [400]])

        np.testing.assert_array_equal(result, expected_result.astype(int))

    @patch('btb.tuning.hyperparams.numerical.sys')
    def test__inverse_transform_min_max(self, mock_sys):
        """Test that the method ``_inverse_transform`` performs a normalization of values between
        ``min`` and ``max`` with a min and max value set up.
        """
        # setup
        _min = 0
        _max = 10
        instance = IntHyperParam(min=_min, max=_max)
        values = np.array([[0.1], [0.9]])

        # run
        result = instance._inverse_transform(values)

        # assert
        expected_result = np.array([[1],
                                    [9]])

        np.testing.assert_array_equal(result, expected_result.astype(int))

    @patch('btb.tuning.hyperparams.numerical.np.random.random')
    @patch('btb.tuning.hyperparams.numerical.IntHyperParam._transform')
    @patch('btb.tuning.hyperparams.numerical.IntHyperParam._inverse_transform')
    def test_sample(self, mock__inverse_transform, mock__transform, mock_np_random):
        """Test that the method ``sample`` returns random generated numbers and process them thro
        the internal methods to convert them in the range of our search space."""
        # setup
        instance = IntHyperParam()
        n_samples = 2

        # run
        result = instance.sample(n_samples)

        # assert
        mock_np_random.assert_called_once_with((n_samples, instance.dimensions))
        mock__inverse_transform.assert_called_once_with(mock_np_random.return_value)
        mock__transform.assert_called_once_with(mock__inverse_transform.return_value)
        self.assertEqual(result, mock__transform.return_value)



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock

import numpy as np

from btb.tuning.metamodels.base import BaseMetaModel


class TestBaseMetaModel(TestCase):
    """Test BaseMetaModel class."""

    def test__init_model__MODEL_KWARGS_DEFAULT_none(self):
        # setup
        instance = MagicMock()
        instance._MODEL_KWARGS_DEFAULT = None
        instance._MODEL_CLASS.return_value = 'model_instance'

        # run
        BaseMetaModel._init_model(instance)

        # assert
        assert instance._model_instance == 'model_instance'
        instance._MODEL_CLASS.assert_called_once_with()

    def test__init_model__MODEL_KWARGS_DEFAULT(self):
        # setup
        instance = MagicMock()
        instance._MODEL_KWARGS_DEFAULT = {'a': 1}
        instance._MODEL_CLASS.return_value = 'model_instance'

        # run
        BaseMetaModel._init_model(instance)

        # assert
        assert instance._model_instance == 'model_instance'
        instance._MODEL_CLASS.assert_called_once_with(a=1)

    def test__init_model__self_model_kwargs(self):
        # setup
        instance = MagicMock()
        instance._model_kwargs = {'b': 2}
        instance._MODEL_KWARGS_DEFAULT = {'a': 1}
        instance._MODEL_CLASS.return_value = 'model_instance'

        # run
        BaseMetaModel._init_model(instance)

        # assert
        assert instance._model_instance == 'model_instance'
        instance._MODEL_CLASS.assert_called_once_with(a=1, b=2)

    def test__fit(self):
        # setup
        instance = MagicMock()

        # run
        BaseMetaModel._fit(instance, 'trials', 'scores')

        # assert
        instance._init_model.called_once_with()
        instance._model_instance.fit.called_once_with('trials', 'scores')

    def test__predict(self):
        # setup
        instance = MagicMock()
        instance._model_instance.predict.return_value = [1]

        # run
        result = BaseMetaModel._predict(instance, 'trials')

        # assert
        np.testing.assert_array_equal(result, np.array([1]))
        instance._model_instance.predict.assert_called_once_with('trials')



# -*- coding: utf-8 -*-

import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF

from btb.tuning.metamodels.base import BaseMetaModel


class GaussianProcessMetaModel(BaseMetaModel):
    """GaussianProcessMetaModel class.

    This class represents a meta-model using an underlying ``GaussianProcessRegressor`` from
    ``sklearn.gaussian_process``.

    Attributes:
        _MODEL_KWARGS (dict):
            Dictionary with the default ``kwargs`` for the ``GaussianProcessRegressor``
            instantiation.
        _MODEL_CLASS (type):
            Class to be instantiated and used for the ``self._model`` instantiation. In
            this case ``sklearn.gaussian_process.GaussainProcessRegressor``
    """
    _MODEL_CLASS = GaussianProcessRegressor

    _MODEL_KWARGS_DEFAULT = {
        'normalize_y': True
    }

    def __init_metamodel__(self, length_scale=1):
        if self._model_kwargs is None:
            self._model_kwargs = {}

        self._model_kwargs['kernel'] = RBF(length_scale=length_scale)

    def _predict(self, candidates):
        predictions = self._model_instance.predict(candidates, return_std=True)
        return np.column_stack(predictions)



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock, call, patch

import numpy as np

from btb.tuning.tunable import Tunable
from btb.tuning.tuners.base import BaseMetaModelTuner, BaseTuner, StopTuning


class TestBaseTuner(TestCase):
    """Test BaseTuner class."""

    def test___init__defaults(self):
        # setup
        tunable = MagicMock(spec_set=Tunable)

        # run
        instance = BaseTuner(tunable)

        # assert
        assert instance.tunable is tunable
        assert isinstance(instance.trials, np.ndarray)
        assert isinstance(instance.raw_scores, np.ndarray)
        assert isinstance(instance.scores, np.ndarray)
        assert isinstance(instance._trials_set, set)
        assert isinstance(instance.maximize, bool)

        assert instance.maximize
        assert instance.trials.shape == (0, 1)
        assert instance.raw_scores.shape == (0, 1)
        assert instance.trials.dtype == np.float
        assert instance.raw_scores.dtype == np.float

    def test___init__maximize_false(self):
        # setup
        tunable = MagicMock(spec_set=Tunable)

        # run
        instance = BaseTuner(tunable, False)

        # assert
        assert isinstance(instance.tunable, MagicMock)
        assert isinstance(instance.trials, np.ndarray)
        assert isinstance(instance.raw_scores, np.ndarray)
        assert isinstance(instance.scores, np.ndarray)
        assert isinstance(instance._trials_set, set)
        assert isinstance(instance.maximize, bool)

        assert not instance.maximize

    def test__check_proposals_proposals_gt_cardinality(self):
        """Test that ``StopTuning`` is being raised if ``proposals`` is greater than
        ``self.tunable.cardinality``.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.cardinality = 4

        # run / assert
        with self.assertRaises(StopTuning):
            BaseTuner._check_proposals(instance, 5)

    def test__check_proposals_trials_eq_cardinality(self):
        """Test that ``StopTuning`` is being raised if ``self.trials`` is equal to
        ``self.tunable.cardinality``.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.cardinality = 2
        instance._trials_set.__len__.return_value = 2

        # run / assert
        with self.assertRaises(StopTuning):
            BaseTuner._check_proposals(instance, 1)

    def test__check_proposals_trials_and_proposals_gt_cardinality(self):
        """Test that ``StopTuning`` is being raised if ``proposals`` + ``len(self.trials)``
        is greater than ``self.tunable.cardinality``.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.cardinality = 4
        instance._trials_set.__len__.return_value = 2

        # run / assert
        with self.assertRaises(StopTuning):
            BaseTuner._check_proposals(instance, 3)

    def test__check_proposals_not_raise(self):
        """Test that ``StopTuning`` is not being raised."""
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.cardinality = 4
        instance._trials_set.__len__.return_value = 2

        # run / assert
        result = BaseTuner._check_proposals(instance, 1)
        assert result is None

    def test_propose_one_value_no_duplicates(self):
        """Test that ``propose`` method calls it's child implemented method with
        ``allow_duplicates`` as ``False``.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        inverse_return = instance.tunable.inverse_transform.return_value
        inverse_return.to_dict.return_value = [1]
        instance._propose = MagicMock(return_value=1)

        # run
        result = BaseTuner.propose(instance, 1)

        # assert
        instance._check_proposals.assert_called_once_with(1)
        instance._propose.assert_called_once_with(1, False)
        instance.tunable.inverse_transform.called_once_with(1)
        inverse_return.to_dict.assert_called_once_with(orient='records')
        assert result == 1

    @patch('btb.tuning.tuners.base.BaseTuner._check_proposals')
    def test_propose_one_value_allow_duplicates(self, mock__check_proposals):
        """Test that ``propose`` method calls it's child implemented method with
        ``allow_duplicates`` as ``True``.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        inverse_return = instance.tunable.inverse_transform.return_value
        inverse_return.to_dict.return_value = [1]
        instance._propose = MagicMock(return_value=1)

        # run
        result = BaseTuner.propose(instance, 1, allow_duplicates=True)

        # assert
        instance._check_proposals.assert_not_called()
        instance._propose.assert_called_once_with(1, True)
        instance.tunable.inverse_transform.called_once_with(1)
        inverse_return.to_dict.assert_called_once_with(orient='records')
        assert result == 1

    @patch('btb.tuning.tuners.base.BaseTuner._check_proposals')
    def test_propose_many_values_no_duplicates(self, mock__check_proposals):
        """Test that ``propose`` method calls it's child implemented method with more than one
        proposals and ``allow_duplicates`` as ``False``.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        inverse_return = instance.tunable.inverse_transform.return_value
        inverse_return.to_dict.return_value = [1, 2]
        instance._propose = MagicMock(return_value=2)

        # run
        result = BaseTuner.propose(instance, 2)

        # assert
        instance._propose.assert_called_once_with(2, False)
        instance.tunable.inverse_transform.called_once_with(2)
        inverse_return.to_dict.assert_called_once_with(orient='records')
        assert result == [1, 2]

    @patch('btb.tuning.tuners.base.BaseTuner._check_proposals')
    def test_propose_many_values_allow_duplicates(self, mock__check_proposals):
        """Test that ``propose`` method calls it's child implemented method with more than one
        proposals and ``allow_duplicates`` as ``True``.
        """

        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        inverse_return = instance.tunable.inverse_transform.return_value
        inverse_return.to_dict.return_value = [1, 2]
        instance._propose = MagicMock(return_value=2)

        # run
        result = BaseTuner.propose(instance, 2, allow_duplicates=True)

        # assert
        instance._propose.assert_called_once_with(2, True)
        instance.tunable.inverse_transform.called_once_with(2)
        inverse_return.to_dict.assert_called_once_with(orient='records')
        assert result == [1, 2]

    def test__sample_allow_duplicates(self):
        """Test the method ``_sample``when using duplicates."""
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.sample.return_value = 1

        # run
        result = BaseTuner._sample(instance, 1, True)

        # assert
        instance.tunable.sample.assert_called_once_with(1)
        assert result == 1

    def test__sample_not_allow_duplicates(self):
        """Test that the method ``_sample`` returns ``np.ndarray`` when not using duplicates."""
        # setup
        instance = MagicMock()
        instance._trials_set = set()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.sample.return_value = np.array([[3]])

        # run
        result = BaseTuner._sample(instance, 1, False)

        # assert
        instance.tunable.sample.assert_called_once_with(1)
        np.testing.assert_array_equal(result, np.array([[3]]))

    def test_sample_no_duplicates_more_than_one_loop(self):
        """Test that the method ``_sample`` returns ``np.ndarray`` when not using duplicates and
        perfroms more than one iteration.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance._trials_set = set({(1, ), (2, )})

        side_effect = [np.array([[3]]), np.array([[1]]), np.array([[1]]), np.array([[4]])]
        instance.tunable.sample.side_effect = side_effect

        # run
        result = BaseTuner._sample(instance, 2, False)

        # assert
        assert instance.tunable.sample.call_args_list == [call(2), call(2), call(2), call(2)]
        np.testing.assert_array_equal(result, np.array([[3], [4]]))

    def test_record_list_maximize_true(self):
        """Test that the method record updates the ``trials``  and ``scores``."""

        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.transform.return_value = np.array([[1, 0]])
        instance.trials = np.empty((0, 2), dtype=np.float)
        instance._trials_set = set()
        instance.scores = None
        instance.maximize = True
        instance.raw_scores = np.empty((0, 1), dtype=np.float)

        # run
        BaseTuner.record(instance, [1], [0.1])

        # assert
        instance.tunable.transform.assert_called_once_with([1])

        np.testing.assert_array_equal(instance.trials, np.array([[1, 0]]))
        assert instance._trials_set == set({(1, 0)})
        np.testing.assert_array_equal(instance.raw_scores, np.array([0.1]))
        np.testing.assert_array_equal(instance.scores, np.array([0.1]))

    def test_record_list_maximize_false(self):
        """Test that the method record updates the ``trials``  and ``scores``."""

        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.transform.return_value = np.array([[1, 0]])
        instance.trials = np.empty((0, 2), dtype=np.float)
        instance._trials_set = set()
        instance.scores = None
        instance.maximize = False
        instance.raw_scores = np.empty((0, 1), dtype=np.float)

        # run
        BaseTuner.record(instance, [1], [0.1])

        # assert
        instance.tunable.transform.assert_called_once_with([1])

        np.testing.assert_array_equal(instance.trials, np.array([[1, 0]]))
        assert instance._trials_set == set({(1, 0)})
        np.testing.assert_array_equal(instance.raw_scores, np.array([0.1]))
        np.testing.assert_array_equal(instance.scores, np.array([-0.1]))

    def test_record_scalar_values(self):
        """Test that the method record performs an update to ``trials`` and ``scores`` when called
        with a scalar value.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.trials = np.empty((0, 2), dtype=np.float)
        instance.raw_scores = np.empty((0, 1), dtype=np.float)
        instance._trials_set = set()
        instance.tunable.transform.return_value = np.array([[1, 0]])

        # run
        BaseTuner.record(instance, 1, 0.1)

        # assert
        instance.tunable.transform.assert_called_once_with(1)
        np.testing.assert_array_equal(instance.trials, np.array([[1, 0]]))
        assert instance._trials_set == set({(1, 0)})
        np.testing.assert_array_equal(instance.raw_scores, np.array([0.1]))
        np.testing.assert_array_equal(instance.scores, np.array([0.1]))

    def test_record_raise_error(self):
        """Test that the method record raises a ``ValueError`` when ``len(trials)`` is different
        to ``len(scores)``.
        """
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.transform.return_value = np.array([[1, 0]])

        # run / assert
        with self.assertRaises(ValueError):
            BaseTuner.record(instance, 1, [1, 2])


class TestBaseMetaModelTuner(TestCase):
    """Test BaseMetaModelTuner class."""

    @patch('btb.tuning.tuners.base.super')
    def test___init___default_values(self, mock_super):
        # setup
        tunable = MagicMock(spec_set=Tunable)
        instance = MagicMock()
        instance.__init_metamodel__ = MagicMock()
        instance.__init_acquisition__ = MagicMock()

        # run
        BaseMetaModelTuner.__init__(instance, tunable)

        # assert
        assert instance.num_candidates == 1000
        assert instance.min_trials == 2
        instance.__init_metamodel__.assert_called_once_with()
        instance.__init_acquisition__.assert_called_once_with()

    @patch('btb.tuning.tuners.base.super')
    def test___init___users_values(self, mock_super):

        # setup
        tunable = MagicMock(spec_set=Tunable)
        instance = MagicMock()
        instance._metamodel_kwargs = {'a': 'test'}
        instance._acquisition_kwargs = {'a': 'acquisition_test'}
        instance.__init_metamodel__ = MagicMock()
        instance.__init_acquisition__ = MagicMock()

        # run
        BaseMetaModelTuner.__init__(
            instance,
            tunable,
            maximize=False,
            num_candidates=5,
            min_trials=20,
        )

        # assert
        assert instance.num_candidates == 5
        assert instance.min_trials == 20
        instance.__init_metamodel__.assert_called_once_with(a='test')
        instance.__init_acquisition__.assert_called_once_with(a='acquisition_test')

    def test__proposemin_trials_gt__trials_set(self):
        # setup
        instance = MagicMock()
        instance.min_trials = 1
        instance._trials_set.__len__.return_value = 0
        instance._sample.return_value = 'sample'

        # run
        result = BaseMetaModelTuner._propose(instance, 1, True)

        # assert
        instance._sample.assert_called_once_with(1, True)
        assert result == 'sample'

    def test__proposemin_trials_lt__trials_set_allow_duplicates(self):
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.cardinality = 3
        instance.min_trials = 0
        instance.num_candidates = 10
        instance._trials_set.__len__.return_value = 1
        instance._sample.return_value = np.array([1])
        instance._predict.return_value = 'predicted'
        instance._acquire.return_value = 0

        # run
        result = BaseMetaModelTuner._propose(instance, 1, True)

        # assert
        instance._sample.assert_called_once_with(10, True)
        instance._predict.assert_called_once_with(np.array([1]))
        assert result == 1

    def test__proposemin_trials_lt__trials_set_not_allow_duplicates(self):
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.cardinality = 3
        instance.min_trials = 0
        instance.num_candidates = 10
        instance._trials_set.__len__.return_value = 1
        instance._sample.return_value = np.array([1])
        instance._predict.return_value = 'predicted'
        instance._acquire.return_value = 0

        # run
        result = BaseMetaModelTuner._propose(instance, 1, False)

        # assert
        instance._sample.assert_called_once_with(2, False)
        instance._predict.assert_called_once_with(np.array([1]))
        instance._acquire.assert_called_once_with('predicted', 1)
        assert result == 1

    def test__proposemin_trials_lt__trials_set_not_allow_duplicates_num_samples_is_min(self):
        # setup
        instance = MagicMock()
        instance.tunable = MagicMock(spec_set=Tunable)
        instance.tunable.cardinality = 3
        instance.min_trials = 0
        instance.num_candidates = 0
        instance._trials_set.__len__.return_value = 1
        instance._sample.return_value = np.array([1])
        instance._predict.return_value = 'predicted'
        instance._acquire.return_value = 0

        # run
        result = BaseMetaModelTuner._propose(instance, 1, False)

        # assert
        instance._sample.assert_called_once_with(0, False)
        instance._predict.assert_called_once_with(np.array([1]))
        instance._acquire.assert_called_once_with('predicted', 1)
        assert result == 1

    @patch('btb.tuning.tuners.base.super')
    def test_record(self, mock_super):
        # setup
        instance = MagicMock()
        instance.trials = np.array([2])
        instance.scores = 1
        instance.min_trials = 1

        # run
        BaseMetaModelTuner.record(instance, 1, 1)

        # assert
        mock_super.return_value.record.assert_called_once_with(1, 1)
        instance._fit.assert_called_once_with(np.array([2]), 1)



# -*- coding: utf-8 -*-

from unittest import TestCase
from unittest.mock import MagicMock, call, patch

import numpy as np

from btb.tuning.hyperparams.categorical import CategoricalHyperParam


def assert_called_with_np_array(mock_calls, real_calls):
    assert len(mock_calls) == len(real_calls)

    for mock_call, real_call in zip(mock_calls, real_calls):
        np.testing.assert_array_equal(mock_call[0], real_call[1])


class TestCategoricalHyperParam(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.instance = CategoricalHyperParam(choices=['Cat', 'Dog', 'Horse', 'Tiger'])

    @patch('btb.tuning.hyperparams.categorical.OneHotEncoder')
    def test___init__(self, mock_one_hot_encoder):
        """Test that during instantiation we create a OneHotEncoder and we fit it with the given
        choices."""
        # setup
        choices = ['cat', 'dog', 'parrot']
        encoder_instance = MagicMock()
        mock_one_hot_encoder.return_value = encoder_instance

        # run
        instance = CategoricalHyperParam(choices=choices)

        # assert
        self.assertEqual(instance.choices, choices)
        self.assertEqual(instance.default, 'cat')

        # TODO: Fix / reimplmement assert_called_with_np
        # expected_encoder_calls = [
        #     call(categories=[np.array(['cat', 'dog', 'parrot'], dtype=object)], sparse=True)
        # ]

        # assert_called_with_np_array(
        #     mock_one_hot_encoder.call_args_list,
        #     expected_encoder_calls
        # )
        expected_encoder_fit_call = [call(np.array(choices).reshape(-1, 1))]
        assert_called_with_np_array(encoder_instance.fit.call_args_list, expected_encoder_fit_call)

    def test__within_hyperparam_space_values_in_space(self):
        """Test that when we call ``_within_hyperparam_space`` with values in the hyperparameter
        search space does not raise an exception."""
        # setup
        values = ['Cat']
        values_2 = ['Dog', 'Horse']

        # run
        self.instance._within_hyperparam_space(values)
        self.instance._within_hyperparam_space(values_2)

    def test__within_hyperparam_space_values_out_of_space(self):
        """Test that when we call ``_within_hyperparam_space`` with values out of the
        hyperparameter search space does not raise an exception."""
        # setup
        values = ['mat']
        values_2 = ['pug', 'Horse']

        # run/assert
        with self.assertRaises(ValueError):
            self.instance._within_hyperparam_space(values)

        with self.assertRaises(ValueError):
            self.instance._within_hyperparam_space(values_2)

    def test__transform_single_value(self):
        """Test that the method ``transform`` performs a normalization over categorical single
        value and converts it in to a search space of [0, 1]^k.
        """

        # setup
        value = np.array([['Cat']])

        # run
        result = self.instance.transform(value)

        # assert
        expected_result = np.array([[1, 0, 0, 0]])

        np.testing.assert_array_equal(result, expected_result)

    def test__transform_multiple_values(self):
        """Test that the method ``transform`` performs a normalization over categorical single
        value and converts it in to a search space of [0, 1]^k.
        """

        # setup
        values = np.array([['Cat'], ['Dog']])

        # run
        results = self.instance.transform(values)

        # assert
        expected_results = np.array([[1, 0, 0, 0], [0, 1, 0, 0]])

        np.testing.assert_array_equal(results, expected_results)

    def test__transform_invalid_value(self):
        """Test that a ``ValueError`` is being raised when a value out of the hyperparameter
        space is given.
        """
        # run / assert
        with self.assertRaises(ValueError):
            self.instance._transform(np.array([['cow']]))

    def test__inverse_transform_single_value(self):
        """Test that the method ``_inverse_transform`` performs a denormalization over the search
        space value to the original categorical hyperparameter space.
        """
        # setup
        value = np.array([[1, 0, 0, 0]])

        # run
        result = self.instance.inverse_transform(value)

        # assert
        np.testing.assert_array_equal(result, np.array([['Cat']]))

    def test__inverse_transform_multiple_values(self):
        """Test that the method ``_inverse_transform`` performs a denormalization over the search
        space values to the original categorical hyperparameter space.
        """
        # setup
        values = np.array([[1, 0, 0, 0], [0, 1, 0, 0]])

        # run
        results = self.instance.inverse_transform(values)

        # assert
        np.testing.assert_array_equal(results, np.array([['Cat'], ['Dog']]))

    @patch('btb.tuning.hyperparams.categorical.np.random.random')
    def test_sample(self, mock_np_random):
        """Test that sample returns values."""
        # setup
        mock_np_random.return_value = np.array([
            [0.5, 0.1, 0.2, 0.6],
            [0.1, 0.9, 0.1, 0.6],
        ])
        n = 2

        # run
        results = self.instance.sample(n)

        # assert
        expected_results = np.array([[0, 0, 0, 1], [0, 1, 0, 0]])

        mock_np_random.assert_called_once_with((n, self.instance.dimensions))

        np.testing.assert_array_equal(results, expected_results)

        self.assertEqual(len(results), n)
